# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Vector type to be exchanged with the DBMS."""

from __future__ import annotations as _

import abc as _abc
import struct as _struct
import sys as _sys
from enum import Enum as _Enum

from . import _typing as _t


if _t.TYPE_CHECKING:
    # "Why?", I hear you ask. Because sphinx of course.
    # This beautiful construct helps sphinx to properly resolve the type hints.
    import numpy as _np
    import pyarrow as _pa
else:
    from ._optional_deps import (
        np as _np,
        pa as _pa,
    )


if False:
    # Ugly work-around to make sphinx understand `@_t.overload`
    import typing as _t  # type: ignore[no-redef]


try:
    from ._rust import vector as _vec_rust
    from ._rust.vector import swap_endian as _swap_endian_unchecked_rust
except ImportError:
    _swap_endian_unchecked_rust = None
    _vec_rust = None


__all__ = [
    "Vector",
    "VectorDType",
    "VectorEndian",
]


_DEFAULT = object()


class Vector:
    r"""
    A class representing a Neo4j vector.

    The constructor accepts various types of data to create a vector.
    Depending on ``data``'s type, further arguments may be required/allowed.
    Examples of valid invocations are::

        Vector([1, 2, 3], "i8")
        Vector(b"\x00\x01\x00\x02", VectorDType.I16)
        Vector(b"\x01\x00\x02\x00", "i16", byteorder="little")
        Vector(numpy.array([1, 2, 3]))
        Vector(pyarrow.array([1, 2, 3]))

    Internally, a vector is stored as a contiguous block of memory
    (:class:`bytes`), containing homogeneous values encoded in big-endian
    order. Support for this feature requires a DBMS supporting Bolt version
    6.0 or later.

    :param data:
        The data from which the vector will be constructed.
        The constructor accepts the following types:

        * ``Iterable[float]``, ``Iterable[int]`` (but not ``bytes`` or
          ``bytearray``):
          Use an iterable of floats or an iterable of ints to construct the
          vector from native Python values.
          The ``dtype`` parameter is required.
          See also: :meth:`.from_native`.
        * ``bytes``, ``bytearray``: Use raw bytes to construct the vector.
          The ``dtype`` parameter is required and ``byteorder`` is optional.
        * ``numpy.ndarray``: Use a numpy array to construct the vector.
          No further parameters are accepted.
          See also: :meth:`.from_numpy`.
        * ``pyarrow.Array``: Use a pyarrow array to construct the vector.
          No further parameters are accepted.
          See also: :meth:`.from_pyarrow`.
    :param dtype: The type of the vector.
        See :class:`.VectorDType` for currently supported inner data types.
        See also :attr:`.dtype`.

        This parameter is required if ``data`` is of type :class:`bytes`,
        :class:`bytearray`, ``Iterable[float]``, or ``Iterable[int]``.
        Otherwise, it must be omitted.
    :param byteorder: The endianness of the input data (default: ``"big"``).
        If ``"little"`` is given, ``neo4j-rust-ext`` or ``numpy`` is used to
        speed up the internal byte flipping (if either package is installed).
        Use :data:`sys.byteorder` if you want to use the system's native
        endianness.

        This parameter is optional if ``data`` is of type ``bytes`` or
        ``bytearray``. Otherwise, it must be omitted.

    :raises ValueError:
        Depending on the type of ``data``:
            * ``Iterable[float]``, ``Iterable[int]`` (excluding byte types):
                * If the dtype is not supported.
            * ``bytes``, ``bytearray``:
                * If the dtype is not supported or data's size is not a
                  multiple of dtype's size.
                * If byteorder is not one of ``"big"`` or ``"little"``.
            * ``numpy.ndarray``:
                * If the dtype is not supported.
                * If the array is not one-dimensional.
            * ``pyarrow.Array``:
                * If the array's type is not supported.
                * If the array contains null values.
    :raises TypeError:
        Depending on the type of ``data``:
            * ``Iterable[float]``, ``Iterable[int]`` (excluding byte types):
                * If data's elements don't match the expected type depending on
                  dtype.
    :raises OverflowError:
        Depending on the type of ``data``:
            * ``Iterable[float]``, ``Iterable[int]`` (excluding byte types):
                * If the value is out of range for the given type.

    .. versionadded: 6.0
    """

    __slots__ = ("__weakref__", "_inner")

    _inner: _InnerVector

    @_t.overload
    def __init__(
        self,
        data: _t.Iterable[float],
        dtype: _T_VectorDTypeFloat,
        /,
    ) -> None: ...

    @_t.overload
    def __init__(
        self,
        data: _t.Iterable[int],
        dtype: _T_VectorDTypeInt,
        /,
    ) -> None: ...

    @_t.overload
    def __init__(
        self,
        data: bytes | bytearray,
        dtype: _T_VectorDType,
        /,
        *,
        byteorder: _T_VectorEndian = "big",
    ) -> None: ...

    @_t.overload
    def __init__(self, data: _np.ndarray, /) -> None: ...

    @_t.overload
    def __init__(self, data: _pa.Array, /) -> None: ...

    def __init__(self, data, *args, **kwargs) -> None:
        if isinstance(data, (bytes, bytearray)):
            self._set_bytes(bytes(data), *args, **kwargs)
        elif _np is not None and isinstance(data, _np.ndarray):
            self._set_numpy(data, *args, **kwargs)
        elif _pa is not None and isinstance(data, _pa.Array):
            self._set_pyarrow(data, *args, **kwargs)
        else:
            self._set_native(data, *args, **kwargs)

    def raw(self, /, *, byteorder: _T_VectorEndian = "big") -> bytes:
        """
        Get the raw bytes of the vector.

        The data is a continuous block of memory, containing an array of the
        vector's data type. The data is stored in big-endian order. Pass
        another byte-order to this method to get the converted data.

        :param byteorder: The endianness the data should be returned in.
            If the data's byte-order needs flipping, this method tries to use
            ``neo4j-rust-ext`` or ``numpy``, if installed, to speed up the
            process. Use :data:`sys.byteorder` if you want to use the system's
            native endianness.

        :returns: The raw bytes of the vector.

        :raises ValueError:
            If byteorder is not one of ``"big"`` or ``"little"``.
        """
        match byteorder:
            case "big":
                return self._inner.data
            case "little":
                return self._inner.data_le
            case _:
                raise ValueError(
                    f"Invalid byteorder: {byteorder!r}. "
                    "Must be 'big' or 'little'."
                )

    def set_raw(
        self,
        data: bytes,
        /,
        *,
        byteorder: _T_VectorEndian = "big",
    ) -> None:
        """
        Set the raw bytes of the vector.

        :param data: The new raw bytes of the vector.
        :param byteorder: The endianness of ``data``.
            The data will always be stored in big-endian order. If passed-in
            byte-order needs flipping, this method tries to use
            ``neo4j-rust-ext`` or ``numpy``, if installed, to speed up the
            process. Use :data:`sys.byteorder` if you want to use the system's
            native endianness.

        :raises ValueError:
          * If data's size is not a multiple of dtype's size.
          * If byteorder is not one of ``"big"`` or ``"little"``.
        :raises TypeError: If the data is not of type bytes.
        """
        match byteorder:
            case "big":
                self._inner.data = data
            case "little":
                self._inner.data_le = data
            case _:
                raise ValueError(
                    f"Invalid byteorder: {byteorder!r}. "
                    "Must be 'big' or 'little'."
                )

    @property
    def dtype(self) -> VectorDType:
        """
        Get the type of the vector.

        :returns: The type of the vector.
        """
        return self._inner.dtype

    def __len__(self) -> int:
        """
        Get the number of elements in the vector.

        :returns: The number of elements in the vector.
        """
        return len(self._inner)

    def __str__(self) -> str:
        return str(self._inner)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}({self.raw()!r}, {self.dtype.value!r})"
        )

    @classmethod
    def from_bytes(
        cls,
        data: bytes,
        dtype: _T_VectorDType,
        /,
        *,
        byteorder: _T_VectorEndian = "big",
    ) -> _t.Self:
        """
        Create a Vector instance from raw bytes.

        :param data: The raw bytes to create the vector from.
        :param dtype: The type of the vector.
            See also :attr:`.dtype`.
        :param byteorder: The endianness of the data.
            If ``"little"``, the bytes in data will be flipped to big-endian.
            If installed, ``neo4j-rust-ext`` or ``numpy`` will be used to speed
            up the byte flipping. Use :data:`sys.byteorder` if you want to use
            the system's native endianness.

        :raises ValueError:
          * If data's size is not a multiple of dtype's size.
          * If byteorder is not one of ``"big"`` or ``"little"``.
        :raises TypeError: If the data is not of type bytes.
        """
        obj = cls.__new__(cls)
        obj._set_bytes(data, dtype, byteorder=byteorder)
        return obj

    def _set_bytes(
        self,
        data: bytes,
        dtype: _T_VectorDType,
        /,
        *,
        byteorder: _T_VectorEndian = "big",
    ) -> None:
        self._inner = _get_type(dtype)(data, byteorder=byteorder)

    @classmethod
    @_t.overload
    def from_native(
        cls, data: _t.Iterable[float], dtype: _T_VectorDTypeFloat, /
    ) -> _t.Self: ...

    @classmethod
    @_t.overload
    def from_native(
        cls, data: _t.Iterable[int], dtype: _T_VectorDTypeInt, /
    ) -> _t.Self: ...

    @classmethod
    def from_native(
        cls,
        data: _t.Iterable[float] | _t.Iterable[int],
        dtype: _T_VectorDType,
        /,
    ) -> _t.Self:
        """
        Create a Vector instance from an iterable of values.

        :param data: The list, tuple, or other iterable of values to create the
            vector from.
        :param dtype: The type of the vector.
            See also :attr:`.dtype`.

        ``data`` must contain values that match the expected type given by
        ``dtype``:

        * ``dtype == "f32"``: :class:`float`
        * ``dtype == "f64"``: :class:`float`
        * ``dtype == "i8"``: :class:`int`
        * ``dtype == "i16"``: :class:`int`
        * ``dtype == "i32"``: :class:`int`
        * ``dtype == "i64"``: :class:`int`

        :raises ValueError: If the dtype is not supported.
        :raises TypeError: If data's elements don't match the expected type
            depending on dtype.
        :raises OverflowError: If the value is out of range for the given type.
        """
        obj = cls.__new__(cls)
        obj._set_native(data, dtype)
        return obj

    def _set_native(
        self,
        data: _t.Iterable[float] | _t.Iterable[int],
        dtype: _T_VectorDType,
        /,
    ) -> None:
        self._inner = _get_type(dtype).from_native(data)

    def to_native(self) -> list[object]:
        """
        Convert the vector to a native Python list.

        The type of the elements in the list depends on the dtype of the
        vector. See :meth:`Vector.from_native` for details.

        :returns: A list of values representing the vector.
        """
        return self._inner.to_native()

    @classmethod
    def from_numpy(cls, data: _np.ndarray, /) -> _t.Self:
        """
        Create a Vector instance from a numpy array.

        :param data: The numpy array to create the vector from.
            The array must be one-dimensional and have a dtype that is
            supported by Neo4j vectors: ``float64``, ``float32``,
            ``int64``, ``int32``, ``int16``, or ``int8``.
            See also :class:`.VectorDType`.

        :raises ValueError:
          * If the dtype is not supported.
          * If the array is not one-dimensional.
        :raises ImportError: If numpy is not installed.

        :returns: A Vector instance constructed from the numpy array.
        """
        obj = cls.__new__(cls)
        obj._set_numpy(data)
        return obj

    def to_numpy(self) -> _np.ndarray:
        """
        Convert the vector to a numpy array.

        The array's dtype depends on the dtype of the vector. However, it will
        always be in big-endian order.

        :returns: A numpy array representing the vector.

        :raises ImportError: If numpy is not installed.
        """
        return self._inner.to_numpy()

    def _set_numpy(self, data: _np.ndarray, /) -> None:
        if data.ndim != 1:
            raise ValueError("Data must be one-dimensional")
        type_: type[_InnerVector]
        match data.dtype.name:
            case "float64":
                type_ = _VecF64
            case "float32":
                type_ = _VecF32
            case "int64":
                type_ = _VecI64
            case "int32":
                type_ = _VecI32
            case "int16":
                type_ = _VecI16
            case "int8":
                type_ = _VecI8
            case _:
                raise ValueError(f"Unsupported numpy dtype: {data.dtype.name}")
        self._inner = type_.from_numpy(data)

    @classmethod
    def from_pyarrow(cls, data: _pa.Array, /) -> _t.Self:
        """
        Create a Vector instance from a pyarrow array.

        PyArrow stores data in little endian. Therefore, the byte-order needs
        to be swapped. If ``neo4j-rust-ext`` or ``numpy`` is installed, it will
        be used to speed up the byte flipping.

        :param data: The pyarrow array to create the vector from.
            The array must have a type that is supported by Neo4j.
            See also :class:`.VectorDType`.
        :raises ValueError:
          * If the array's type is not supported.
          * If the array contains null values.
        :raises ImportError: If pyarrow is not installed.

        :returns: A Vector instance constructed from the pyarrow array.
        """
        obj = cls.__new__(cls)
        obj._set_pyarrow(data)
        return obj

    def to_pyarrow(self) -> _pa.Array:
        """
        Convert the vector to a pyarrow array.

        :returns: A pyarrow array representing the vector.

        :raises ImportError: If pyarrow is not installed.
        """
        return self._inner.to_pyarrow()

    def _set_pyarrow(self, data: _pa.Array, /) -> None:
        import pyarrow

        type_: type[_InnerVector]
        if data.type == pyarrow.float64():
            type_ = _VecF64
        elif data.type == pyarrow.float32():
            type_ = _VecF32
        elif data.type == pyarrow.int64():
            type_ = _VecI64
        elif data.type == pyarrow.int32():
            type_ = _VecI32
        elif data.type == pyarrow.int16():
            type_ = _VecI16
        elif data.type == pyarrow.int8():
            type_ = _VecI8
        else:
            raise ValueError(f"Unsupported pyarrow dtype: {data.type}")
        inner = type_.from_pyarrow(data)
        self._inner = inner

    # TODO: consider conversion to/from
    #   * tensorflow
    #   * pandas
    #   * polars


class VectorEndian(str, _Enum):
    """
    Data endianness (i.e., byte order) of the elements in a :class:`Vector`.

    Inherits from :class:`str` and :class:`enum.Enum`.
    Every driver API accepting a :class:`.VectorEndian` value will also accept
    a string::

        >>> VectorEndian.BIG == "big"
        True
        >>> VectorEndian.LITTLE == "little"
        True

    .. seealso:: :attr:`Vector.raw`

    .. versionadded:: 6.0
    """

    BIG = "big"
    LITTLE = "little"


_T_VectorEndian = VectorEndian | _t.Literal["big", "little"]


class VectorDType(str, _Enum):
    """
    The data type of the elements in a :class:`Vector`.

    Currently supported types are:

        * ``f32``: 32-bit floating point number (single)
        * ``f64``: 64-bit floating point number (double)
        * ``i8``: 8-bit integer
        * ``i16``: 16-bit integer
        * ``i32``: 32-bit integer
        * ``i64``: 64-bit integer

    Inherits from :class:`str` and :class:`enum.Enum`.
    Every driver API accepting a :class:`.VectorDType` value will also accept
    a string::

        >>> VectorDType.F32 == "f32"
        True
        >>> VectorDType.I8 == "i8"
        True

    .. seealso:: :attr:`Vector.dtype`

    .. versionadded:: 6.0
    """

    F32 = "f32"
    F64 = "f64"
    I8 = "i8"
    I16 = "i16"
    I32 = "i32"
    I64 = "i64"


_T_VectorDType = (
    VectorDType | _t.Literal["f32", "f64", "i8", "i16", "i32", "i64"]
)
_T_VectorDTypeInt = _t.Literal[
    VectorDType.I8,
    VectorDType.I16,
    VectorDType.I32,
    VectorDType.I64,
    "i8",
    "i16",
    "i32",
    "i64",
]
_T_VectorDTypeFloat = _t.Literal[
    VectorDType.F32, VectorDType.F64, "f32", "f64"
]


def _swap_endian(type_size: int, data: bytes, /) -> bytes:
    """Swap from big endian to little endian."""
    if type_size == 1:
        return data
    if type_size not in {2, 4, 8}:
        raise ValueError(f"Unsupported type size: {type_size}")
    if len(data) % type_size != 0:
        raise ValueError(
            f"Data length {len(data)} is not a multiple of {type_size}"
        )
    return _swap_endian_unchecked(type_size, data)


def _swap_endian_unchecked_np(type_size: int, data: bytes, /) -> bytes:
    dtype: _np.dtype
    match type_size:
        case 2:
            dtype = _np.dtype("<i2")
        case 4:
            dtype = _np.dtype("<i4")
        case 8:
            dtype = _np.dtype("<i8")
        case _:
            raise ValueError(f"Unsupported type size: {type_size}")
    return _np.frombuffer(data, dtype=dtype).byteswap().tobytes()


def _swap_endian_unchecked_py(type_size: int, data: bytes, /) -> bytes:
    return bytes(
        byte
        for i in range(0, len(data), type_size)
        for byte in data[i : i + type_size][::-1]
    )


if _swap_endian_unchecked_rust is not None:
    _swap_endian_unchecked = _swap_endian_unchecked_rust
elif _np is not None:
    _swap_endian_unchecked = _swap_endian_unchecked_np
else:
    _swap_endian_unchecked = _swap_endian_unchecked_py


def _get_type(dtype: _T_VectorDType, /) -> type[_InnerVector]:
    if isinstance(dtype, str):
        if dtype not in VectorDType.__members__.values():
            raise ValueError(f"Unsupported vector type: {dtype!r}.")
        dtype = VectorDType(dtype)
    if not isinstance(dtype, VectorDType):
        raise TypeError(f"Expected a VectorDType or str, got {type(dtype)}.")
    if dtype not in _TYPES:
        raise ValueError(f"Unsupported vector type: {dtype!r}.")
    return _TYPES[dtype]


_TYPES: dict[VectorDType, type[_InnerVector]] = {}


class _InnerVector(_abc.ABC):
    __slots__ = ("_data", "_data_le")

    dtype: _t.ClassVar[VectorDType]
    size: _t.ClassVar[int]
    cypher_inner_type_repr: _t.ClassVar[str]
    _data: bytes
    _data_le: bytes | None

    def __init__(
        self, data: bytes, /, *, byteorder: _T_VectorEndian = "big"
    ) -> None:
        super().__init__()
        if self.__class__ == _InnerVector:
            raise TypeError("Cannot instantiate abstract class InnerVector")
        match byteorder:
            case "big":
                self.data = data
                self._data_le = None
            case "little":
                self.data = _swap_endian(self.size, data)
                self._data_le = data
            case _:
                raise ValueError(
                    f"Invalid byteorder: {byteorder!r}. "
                    "Must be 'big' or 'little'."
                )

    @property
    def data(self) -> bytes:
        return self._data

    @data.setter
    def data(self, data: bytes, /) -> None:
        if not isinstance(data, bytes):
            raise TypeError("Data must be of type bytes")
        if not len(data) % self.size == 0:
            raise ValueError(
                f"Data length {len(data)} is not a multiple of {self.size}"
            )
        self._data = data

    @property
    def data_le(self) -> bytes:
        if self._data_le is None:
            self._data_le = _swap_endian(self.size, self.data)
        return self._data_le

    @data_le.setter
    def data_le(self, data: bytes, /) -> None:
        self.data = _swap_endian(self.size, data)
        self._data_le = data

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        if _abc.ABC in cls.__bases__:
            return
        dtype = getattr(cls, "dtype", None)
        if not isinstance(dtype, VectorDType):
            raise TypeError(
                f"Class {cls.__name__} must have a VectorDType attribute"
                "'dtype'"
            )
        if not isinstance(getattr(cls, "size", None), int):
            raise TypeError(
                f"Class {cls.__name__} must have a str attribute 'size'"
            )
        if cls.size not in {1, 2, 4, 8}:
            # Either change the sub-type's size if it was a typo or add support
            # for the new size in the swap_endian function.
            raise ValueError(
                f"Class {cls.__name__} has an unhandled size {cls.size}"
            )
        if dtype in _TYPES:
            raise ValueError(
                f"Class {cls.__name__} has a duplicate type '{dtype}'"
            )
        _TYPES[dtype] = cls

    def __len__(self) -> int:
        return len(self.data) // self.size

    def __str__(self) -> str:
        size = len(self)
        type_repr = self.cypher_inner_type_repr
        values_repr = self._cypher_values_repr()
        return f"vector({values_repr}, {size}, {type_repr})"

    @_abc.abstractmethod
    def _cypher_values_repr(self) -> str: ...

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        return f"{cls_name}({self.data!r})"

    @classmethod
    @_abc.abstractmethod
    def from_native(cls, data: _t.Iterable[object], /) -> _t.Self: ...

    @_abc.abstractmethod
    def to_native(self) -> list[object]: ...

    @classmethod
    def from_numpy(cls, data: _np.ndarray, /) -> _t.Self:
        if data.dtype.byteorder == "<" or (
            data.dtype.byteorder == "=" and _sys.byteorder == "little"
        ):
            data = data.byteswap()
        return cls(data.tobytes())

    @_abc.abstractmethod
    def to_numpy(self) -> _np.ndarray: ...

    @classmethod
    def from_pyarrow(cls, data: _pa.Array, /) -> _t.Self:
        width = data.type.byte_width
        assert cls.size == width
        if _pa.compute.count(data, mode="only_null").as_py():
            raise ValueError("PyArrow array must not contain any null values.")
        _, buffer = data.buffers()
        buffer = buffer[
            data.offset * width : (data.offset + len(data)) * width
        ]
        return cls(bytes(buffer), byteorder=_sys.byteorder)

    @_abc.abstractmethod
    def to_pyarrow(self) -> _pa.Array: ...


class _InnerVectorFloat(_InnerVector, _abc.ABC):
    __slots__ = ()

    def _cypher_values_repr(self) -> str:
        res = str(self.to_native())
        return res.replace("nan", "NaN").replace("inf", "Infinity")


class _VecF64(_InnerVectorFloat):
    __slots__ = ()

    dtype = VectorDType.F64
    size = 8
    cypher_inner_type_repr = "FLOAT NOT NULL"

    @classmethod
    def _from_native_rust(cls, data: _t.Iterable[object], /) -> _t.Self:
        return cls(_vec_rust.vec_f64_from_native(data))

    @classmethod
    def _from_native_np(cls, data: _t.Iterable[object], /) -> _t.Self:
        data = tuple(data)
        non_float_gen = (item for item in data if not isinstance(item, float))
        non_float = next(non_float_gen, _DEFAULT)
        if non_float is not _DEFAULT:
            raise TypeError(
                f"Cannot build f64 vector from {type(non_float).__name__}, "
                "expected float."
            )
        return cls(_np.fromiter(data, dtype=_np.dtype(">f8")).tobytes())

    @classmethod
    def _from_native_py(cls, data: _t.Iterable[object], /) -> _t.Self:
        bytes_ = bytearray()
        for item in data:
            if not isinstance(item, float):
                raise TypeError(
                    f"Cannot build f64 vector from {type(item).__name__}, "
                    "expected float."
                )
            bytes_.extend(_struct.pack(">d", item))
        return cls(bytes(bytes_))

    if _vec_rust is not None:
        from_native = _from_native_rust
    elif _np is not None:
        from_native = _from_native_np
    else:
        from_native = _from_native_py

    def _to_native_rust(self) -> list[object]:
        return _vec_rust.vec_f64_to_native(self.data)

    def _to_native_np(self) -> list[object]:
        return _np.frombuffer(self.data, dtype=_np.dtype(">f8")).tolist()

    def _to_native_py(self) -> list[object]:
        return [
            _struct.unpack(">d", self.data[i : i + self.size])[0]
            for i in range(0, len(self.data), self.size)
        ]

    if _vec_rust is not None:
        to_native = _to_native_rust
    elif _np is not None:
        to_native = _to_native_np
    else:
        to_native = _to_native_py

    def to_numpy(self) -> _np.ndarray:
        import numpy

        return numpy.frombuffer(self.data, dtype=numpy.dtype(">f8"))

    def to_pyarrow(self) -> _pa.Array:
        import pyarrow

        buffer = pyarrow.py_buffer(self.data_le)
        return pyarrow.Array.from_buffers(
            pyarrow.float64(), len(self), [None, buffer], 0
        )


class _VecF32(_InnerVectorFloat):
    __slots__ = ()

    dtype = VectorDType.F32
    size = 4
    cypher_inner_type_repr = "FLOAT32 NOT NULL"

    @classmethod
    def _from_native_rust(cls, data: _t.Iterable[object], /) -> _t.Self:
        return cls(_vec_rust.vec_f32_from_native(data))

    @classmethod
    def _from_native_np(cls, data: _t.Iterable[object], /) -> _t.Self:
        data = tuple(data)
        non_float_gen = (item for item in data if not isinstance(item, float))
        non_float = next(non_float_gen, _DEFAULT)
        if non_float is not _DEFAULT:
            raise TypeError(
                f"Cannot build f32 vector from {type(non_float).__name__}, "
                "expected float."
            )
        return cls(_np.fromiter(data, dtype=_np.dtype(">f4")).tobytes())

    @classmethod
    def _from_native_py(cls, data: _t.Iterable[object], /) -> _t.Self:
        bytes_ = bytearray()
        for item in data:
            if not isinstance(item, float):
                raise TypeError(
                    f"Cannot build f32 vector from {type(item).__name__}, "
                    "expected float."
                )
            bytes_.extend(_struct.pack(">f", item))
        return cls(bytes(bytes_))

    if _vec_rust is not None:
        from_native = _from_native_rust
    elif _np is not None:
        from_native = _from_native_np
    else:
        from_native = _from_native_py

    def _to_native_rust(self) -> list[object]:
        return _vec_rust.vec_f32_to_native(self.data)

    def _to_native_np(self) -> list[object]:
        return _np.frombuffer(self.data, dtype=_np.dtype(">f4")).tolist()

    def _to_native_py(self) -> list[object]:
        return [
            _struct.unpack(">f", self.data[i : i + self.size])[0]
            for i in range(0, len(self.data), self.size)
        ]

    if _vec_rust is not None:
        to_native = _to_native_rust
    elif _np is not None:
        to_native = _to_native_np
    else:
        to_native = _to_native_py

    def to_numpy(self) -> _np.ndarray:
        import numpy

        return numpy.frombuffer(self.data, dtype=numpy.dtype(">f4"))

    def to_pyarrow(self) -> _pa.Array:
        import pyarrow

        buffer = pyarrow.py_buffer(self.data_le)
        return pyarrow.Array.from_buffers(
            pyarrow.float32(), len(self), [None, buffer], 0
        )


class _InnerVectorInt(_InnerVector, _abc.ABC):
    __slots__ = ()

    def _cypher_values_repr(self) -> str:
        return str(self.to_native())


_I64_MIN = -9_223_372_036_854_775_808
_I64_MAX = 9_223_372_036_854_775_807


class _VecI64(_InnerVectorInt):
    __slots__ = ()

    dtype = VectorDType.I64
    size = 8
    cypher_inner_type_repr = "INTEGER NOT NULL"

    @classmethod
    def _from_native_rust(cls, data: _t.Iterable[object], /) -> _t.Self:
        return cls(_vec_rust.vec_i64_from_native(data))

    @classmethod
    def _from_native_np(cls, data: _t.Iterable[object], /) -> _t.Self:
        data = tuple(data)
        non_int_gen = (item for item in data if not isinstance(item, int))
        non_int = next(non_int_gen, _DEFAULT)
        if non_int is not _DEFAULT:
            raise TypeError(
                f"Cannot build i64 vector from {type(non_int).__name__}, "
                "expected int."
            )
        data = _t.cast(tuple[int, ...], data)
        overflow_int = tuple(
            item for item in data if not _I64_MIN <= item <= _I64_MAX
        )
        if overflow_int:
            raise OverflowError(
                f"Value {overflow_int[0]} is out of range for i64: "
                f"[-{_I64_MIN}, {_I64_MAX}]"
            )
        return cls(_np.fromiter(data, dtype=_np.dtype(">i8")).tobytes())

    @classmethod
    def _from_native_py(cls, data: _t.Iterable[object], /) -> _t.Self:
        bytes_ = bytearray()
        for item in data:
            if not isinstance(item, int):
                raise TypeError(
                    f"Cannot build i64 vector from {type(item).__name__}, "
                    "expected int."
                )
            if not _I64_MIN <= item <= _I64_MAX:
                raise OverflowError(
                    f"Value {item} is out of range for i64: "
                    f"[-{_I64_MIN}, {_I64_MAX}]"
                )
            bytes_.extend(_struct.pack(">q", item))
        return cls(bytes(bytes_))

    if _vec_rust is not None:
        from_native = _from_native_rust
    elif _np is not None:
        from_native = _from_native_np
    else:
        from_native = _from_native_py

    def _to_native_rust(self) -> list[object]:
        return _vec_rust.vec_i64_to_native(self.data)

    def _to_native_np(self) -> list[object]:
        return _np.frombuffer(self.data, dtype=_np.dtype(">i8")).tolist()

    def _to_native_py(self) -> list[object]:
        return [
            _struct.unpack(">q", self.data[i : i + self.size])[0]
            for i in range(0, len(self.data), self.size)
        ]

    if _vec_rust is not None:
        to_native = _to_native_rust
    elif _np is not None:
        to_native = _to_native_np
    else:
        to_native = _to_native_py

    def to_numpy(self) -> _np.ndarray:
        import numpy

        return numpy.frombuffer(self.data, dtype=numpy.dtype(">i8"))

    def to_pyarrow(self) -> _pa.Array:
        import pyarrow

        buffer = pyarrow.py_buffer(self.data_le)
        return pyarrow.Array.from_buffers(
            pyarrow.int64(), len(self), [None, buffer], 0
        )


_I32_MIN = -2_147_483_648
_I32_MAX = 2_147_483_647


class _VecI32(_InnerVectorInt):
    __slots__ = ()

    dtype = VectorDType.I32
    size = 4
    cypher_inner_type_repr = "INTEGER32 NOT NULL"

    @classmethod
    def _from_native_rust(cls, data: _t.Iterable[object], /) -> _t.Self:
        return cls(_vec_rust.vec_i32_from_native(data))

    @classmethod
    def _from_native_np(cls, data: _t.Iterable[object], /) -> _t.Self:
        data = tuple(data)
        non_int_gen = (item for item in data if not isinstance(item, int))
        non_int = next(non_int_gen, _DEFAULT)
        if non_int is not _DEFAULT:
            raise TypeError(
                f"Cannot build i32 vector from {type(non_int).__name__}, "
                "expected int."
            )
        data = _t.cast(tuple[int, ...], data)
        overflow_int = tuple(
            item for item in data if not _I32_MIN <= item <= _I32_MAX
        )
        if overflow_int:
            raise OverflowError(
                f"Value {overflow_int[0]} is out of range for i32: "
                f"[-{_I32_MIN}, {_I32_MAX}]"
            )
        return cls(_np.fromiter(data, dtype=_np.dtype(">i4")).tobytes())

    @classmethod
    def _from_native_py(cls, data: _t.Iterable[object], /) -> _t.Self:
        bytes_ = bytearray()
        for item in data:
            if not isinstance(item, int):
                raise TypeError(
                    f"Cannot build i32 vector from {type(item).__name__}, "
                    "expected int."
                )
            if not _I32_MIN <= item <= _I32_MAX:
                raise OverflowError(
                    f"Value {item} is out of range for i32: "
                    f"[-{_I32_MIN}, {_I32_MAX}]"
                )
            bytes_.extend(_struct.pack(">i", item))
        return cls(bytes(bytes_))

    if _vec_rust is not None:
        from_native = _from_native_rust
    elif _np is not None:
        from_native = _from_native_np
    else:
        from_native = _from_native_py

    def _to_native_rust(self) -> list[object]:
        return _vec_rust.vec_i32_to_native(self.data)

    def _to_native_np(self) -> list[object]:
        return _np.frombuffer(self.data, dtype=_np.dtype(">i4")).tolist()

    def _to_native_py(self) -> list[object]:
        return [
            _struct.unpack(">i", self.data[i : i + self.size])[0]
            for i in range(0, len(self.data), self.size)
        ]

    if _vec_rust is not None:
        to_native = _to_native_rust
    elif _np is not None:
        to_native = _to_native_np
    else:
        to_native = _to_native_py

    def to_numpy(self) -> _np.ndarray:
        import numpy

        return numpy.frombuffer(self.data, dtype=numpy.dtype(">i4"))

    def to_pyarrow(self) -> _pa.Array:
        import pyarrow

        buffer = pyarrow.py_buffer(self.data_le)
        return pyarrow.Array.from_buffers(
            pyarrow.int32(), len(self), [None, buffer], 0
        )


_I16_MIN = -32_768
_I16_MAX = 32_767


class _VecI16(_InnerVectorInt):
    __slots__ = ()

    dtype = VectorDType.I16
    size = 2
    cypher_inner_type_repr = "INTEGER16 NOT NULL"

    @classmethod
    def _from_native_rust(cls, data: _t.Iterable[object], /) -> _t.Self:
        return cls(_vec_rust.vec_i16_from_native(data))

    @classmethod
    def _from_native_np(cls, data: _t.Iterable[object], /) -> _t.Self:
        data = tuple(data)
        non_int_gen = (item for item in data if not isinstance(item, int))
        non_int = next(non_int_gen, _DEFAULT)
        if non_int is not _DEFAULT:
            raise TypeError(
                f"Cannot build i16 vector from {type(non_int).__name__}, "
                "expected int."
            )
        data = _t.cast(tuple[int, ...], data)
        overflow_int = tuple(
            item for item in data if not _I16_MIN <= item <= _I16_MAX
        )
        if overflow_int:
            raise OverflowError(
                f"Value {overflow_int[0]} is out of range for i16: "
                f"[-{_I16_MIN}, {_I16_MAX}]"
            )
        return cls(_np.fromiter(data, dtype=_np.dtype(">i2")).tobytes())

    @classmethod
    def _from_native_py(cls, data: _t.Iterable[object], /) -> _t.Self:
        bytes_ = bytearray()
        for item in data:
            if not isinstance(item, int):
                raise TypeError(
                    f"Cannot build i16 vector from {type(item).__name__}, "
                    "expected int."
                )
            if not _I16_MIN <= item <= _I16_MAX:
                raise OverflowError(
                    f"Value {item} is out of range for i16: "
                    f"[-{_I16_MIN}, {_I16_MAX}]"
                )
            bytes_.extend(_struct.pack(">h", item))
        return cls(bytes(bytes_))

    if _vec_rust is not None:
        from_native = _from_native_rust
    elif _np is not None:
        from_native = _from_native_np
    else:
        from_native = _from_native_py

    def _to_native_rust(self) -> list[object]:
        return _vec_rust.vec_i16_to_native(self.data)

    def _to_native_np(self) -> list[object]:
        return _np.frombuffer(self.data, dtype=_np.dtype(">i2")).tolist()

    def _to_native_py(self) -> list[object]:
        return [
            _struct.unpack(">h", self.data[i : i + self.size])[0]
            for i in range(0, len(self.data), self.size)
        ]

    if _vec_rust is not None:
        to_native = _to_native_rust
    elif _np is not None:
        to_native = _to_native_np
    else:
        to_native = _to_native_py

    def to_numpy(self) -> _np.ndarray:
        import numpy

        return numpy.frombuffer(self.data, dtype=numpy.dtype(">i2"))

    def to_pyarrow(self) -> _pa.Array:
        import pyarrow

        buffer = pyarrow.py_buffer(self.data_le)
        return pyarrow.Array.from_buffers(
            pyarrow.int16(), len(self), [None, buffer], 0
        )


_I8_MIN = -128
_I8_MAX = 127


class _VecI8(_InnerVectorInt):
    __slots__ = ()

    dtype = VectorDType.I8
    size = 1
    cypher_inner_type_repr = "INTEGER8 NOT NULL"

    @classmethod
    def _from_native_rust(cls, data: _t.Iterable[object], /) -> _t.Self:
        return cls(_vec_rust.vec_i8_from_native(data))

    @classmethod
    def _from_native_np(cls, data: _t.Iterable[object], /) -> _t.Self:
        data = tuple(data)
        non_int_gen = (item for item in data if not isinstance(item, int))
        non_int = next(non_int_gen, _DEFAULT)
        if non_int is not _DEFAULT:
            raise TypeError(
                f"Cannot build i8 vector from {type(non_int).__name__}, "
                "expected int."
            )
        data = _t.cast(tuple[int, ...], data)
        overflow_int = tuple(
            item for item in data if not _I8_MIN <= item <= _I8_MAX
        )
        if overflow_int:
            raise OverflowError(
                f"Value {overflow_int[0]} is out of range for i8: "
                f"[-{_I8_MIN}, {_I8_MAX}]"
            )
        return cls(_np.fromiter(data, dtype=_np.dtype(">i1")).tobytes())

    @classmethod
    def _from_native_py(cls, data: _t.Iterable[object], /) -> _t.Self:
        bytes_ = bytearray()
        for item in data:
            if not isinstance(item, int):
                raise TypeError(
                    f"Cannot build i8 vector from {type(item).__name__}, "
                    "expected int."
                )
            if not _I8_MIN <= item <= _I8_MAX:
                raise OverflowError(
                    f"Value {item} is out of range for i8: "
                    f"[-{_I8_MIN}, {_I8_MAX}]"
                )
            bytes_.extend(_struct.pack(">b", item))
        return cls(bytes(bytes_))

    if _vec_rust is not None:
        from_native = _from_native_rust
    elif _np is not None:
        from_native = _from_native_np
    else:
        from_native = _from_native_py

    def _to_native_rust(self) -> list[object]:
        return _vec_rust.vec_i8_to_native(self.data)

    def _to_native_np(self) -> list[object]:
        return _np.frombuffer(self.data, dtype=_np.dtype(">i1")).tolist()

    def _to_native_py(self) -> list[object]:
        return [
            _struct.unpack(">b", self.data[i : i + self.size])[0]
            for i in range(0, len(self.data), self.size)
        ]

    if _vec_rust is not None:
        to_native = _to_native_rust
    elif _np is not None:
        to_native = _to_native_np
    else:
        to_native = _to_native_py

    def to_numpy(self) -> _np.ndarray:
        import numpy

        return numpy.frombuffer(self.data, dtype=numpy.dtype(">i1"))

    def to_pyarrow(self) -> _pa.Array:
        import pyarrow

        buffer = pyarrow.py_buffer(self.data_le)
        return pyarrow.Array.from_buffers(
            pyarrow.int8(), len(self), [None, buffer], 0
        )



# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

import itertools
import typing as t

# ignore TC003 to make sphinx not completely drop the ball
from collections.abc import Sequence  # noqa: TC003
from copy import deepcopy
from dataclasses import dataclass

from .._api import (
    NotificationCategory,
    NotificationClassification,
    NotificationSeverity,
)
from .._exceptions import BoltProtocolError


if t.TYPE_CHECKING:
    import typing_extensions as te
    from typing_extensions import deprecated

    from .._addressing import Address
    from ..api import ServerInfo

    _T = t.TypeVar("_T")
else:
    from .._warnings import deprecated


class ResultSummary:
    """A summary of execution returned with a :class:`.Result` object."""

    #: A :class:`neo4j.ServerInfo` instance. Provides some basic information of
    #: the server where the result is obtained from.
    server: ServerInfo

    #: The database name where this summary is obtained from.
    database: str | None

    #: The query that was executed to produce this result.
    query: str | None

    #: Dictionary of parameters passed with the statement.
    parameters: dict[str, t.Any] | None

    #: A string that describes the type of query
    #: ``'r'`` = read-only, ``'rw'`` = read/write, ``'w'`` = write-only,
    #: ``'s'`` = schema.
    query_type: t.Literal["r", "rw", "w", "s"] | None

    #: A :class:`neo4j.SummaryCounters` instance. Counters for operations the
    #: query triggered.
    counters: SummaryCounters

    #: Dictionary that describes how the database will execute the query.
    plan: dict | None

    #: Dictionary that describes how the database executed the query.
    profile: dict | None

    #: The time it took for the server to have the result available.
    #: (milliseconds)
    result_available_after: int | None

    #: The time it took for the server to consume the result. (milliseconds)
    result_consumed_after: int | None

    #: see :attr:`.notifications`
    _notifications: list[dict] | None

    # cache for property ``summary_notifications``
    _summary_notifications: tuple[SummaryNotification, ...]

    # cache for property ``summary_notifications``
    _gql_status_objects: tuple[GqlStatusObject, ...]

    _had_key: bool
    _had_record: bool

    def __init__(
        self,
        address: Address,
        had_key: bool,
        had_record: bool,
        metadata: dict[str, t.Any],
    ) -> None:
        self._had_key = had_key
        self._had_record = had_record
        self.metadata = metadata
        self.server = metadata["server"]
        self.database = metadata.get("db")
        self.query = metadata.get("query")
        self.parameters = metadata.get("parameters")
        if "type" in metadata:
            self.query_type = metadata["type"]
            if self.query_type not in {"r", "w", "rw", "s"}:
                raise BoltProtocolError(
                    f"Unexpected query type {self.query_type!r} received from "
                    "server. Consider updating the driver.",
                    address,
                )
        self.query_type = metadata.get("type")
        self.plan = metadata.get("plan")
        self.profile = metadata.get("profile")
        self.counters = SummaryCounters(metadata.get("stats", {}))
        if self.server.protocol_version[0] < 3:
            self.result_available_after = metadata.get(
                "result_available_after"
            )
            self.result_consumed_after = metadata.get("result_consumed_after")
        else:
            self.result_available_after = metadata.get("t_first")
            self.result_consumed_after = metadata.get("t_last")

    @staticmethod
    def _notification_from_status(status: dict) -> dict:
        notification = {}
        for notification_key, status_key in (
            ("title", "title"),
            ("code", "neo4j_code"),
            ("description", "description"),
        ):
            if status_key in status:
                notification[notification_key] = status[status_key]

        if "diagnostic_record" in status:
            diagnostic_record = status["diagnostic_record"]
            if not isinstance(diagnostic_record, dict):
                diagnostic_record = {}

            for notification_key, diag_record_key in (
                ("severity", "_severity"),
                ("category", "_classification"),
                ("position", "_position"),
            ):
                if diag_record_key in diagnostic_record:
                    notification[notification_key] = diagnostic_record[
                        diag_record_key
                    ]

        return notification

    @property
    @deprecated(
        "ResultSummary.notifications is deprecated, "
        "use ResultSummary.gql_status_objects instead."
    )
    def notifications(self) -> list[dict] | None:
        """
        A list of Dictionaries containing notification information.

        Notifications provide extra information for a user executing a
        statement.
        They can be warnings about problematic queries or other valuable
        information that can be presented in a client.
        Unlike failures or errors, notifications do not affect the execution of
        a statement.

        .. seealso:: :attr:`.summary_notifications`

        .. deprecated:: 6.0
            Use :attr:`.gql_status_objects` instead.
        """
        return self._get_notifications()

    @notifications.setter
    @deprecated(
        "ResultSummary.notifications is deprecated, "
        "use ResultSummary.gql_status_objects instead."
    )
    def notifications(self, value: list[dict] | None) -> None:
        self._notifications = value

    def _get_notifications(self) -> list[dict] | None:
        if not hasattr(self, "_notifications"):
            self._set_notifications()
        return self._notifications

    def _set_notifications(self) -> None:
        if "notifications" in self.metadata:
            notifications = self.metadata["notifications"]
            if not isinstance(notifications, list):
                self._notifications = None
                return
            self._notifications = notifications
            return

        # polyfill notifications from GqlStatusObjects
        if "statuses" in self.metadata:
            statuses = self.metadata["statuses"]
            if not isinstance(statuses, list):
                self._notifications = None
                return
            notifications = []
            for status in statuses:
                if not (isinstance(status, dict) and "neo4j_code" in status):
                    # not a notification status
                    continue
                notification = self._notification_from_status(status)
                notifications.append(notification)
            self._notifications = notifications or None
            return

        self._notifications = None

    @property
    @deprecated(
        "ResultSummary.summary_notifications is deprecated, "
        "use ResultSummary.gql_status_objects instead."
    )
    def summary_notifications(self) -> Sequence[SummaryNotification]:
        """
        The same as ``notifications`` but in a parsed, structured form.

        Further, if connected to a gql-aware server, this property will be
        polyfilled from :attr:`gql_status_objects`.

        .. seealso:: :attr:`.notifications`, :class:`.SummaryNotification`

        .. versionadded:: 5.7

        .. deprecated:: 6.0
            Use :attr:`.gql_status_objects` instead.
        """
        if getattr(self, "_summary_notifications", None) is not None:
            return self._summary_notifications

        raw_notifications = self._get_notifications()
        if not isinstance(raw_notifications, list):
            self._summary_notifications = ()
            return self._summary_notifications
        self._summary_notifications = tuple(
            SummaryNotification._from_metadata(n) for n in raw_notifications
        )
        return self._summary_notifications

    @property
    def gql_status_objects(self) -> t.Sequence[GqlStatusObject]:
        """
        Get GqlStatusObjects that arose when executing the query.

        The sequence always contains at least 1 status representing the
        Success, No Data or Omitted Result.
        All other status are notifications like warnings about problematic
        queries or other valuable information that can be presented in a
        client.

        The GqlStatusObjects will be presented in the following order:

        * A "no data" (``02xxx``) has precedence over a warning.
        * A "warning" (``01xxx``) has precedence over a success.
        * A "success" (``00xxx``) has precedence over anything informational
          (``03xxx``).

        .. versionadded:: 5.22

        .. versionchanged:: 6.0 Stabilized from preview.
        """
        raw_status_objects = self.metadata.get("statuses")
        if isinstance(raw_status_objects, list):
            self._gql_status_objects = tuple(
                GqlStatusObject._from_status_metadata(s)
                for s in raw_status_objects
            )
            return self._gql_status_objects

        raw_notifications = self._get_notifications()
        notification_status_objects: t.Iterable[GqlStatusObject]
        if isinstance(raw_notifications, list):
            notification_status_objects = [
                GqlStatusObject._from_notification_metadata(n)
                for n in raw_notifications
            ]
        else:
            notification_status_objects = ()

        if self._had_record:
            # polyfill with a Success status
            result_status = GqlStatusObject._success()
        elif self._had_key:
            # polyfill with an Omitted Result status
            result_status = GqlStatusObject._no_data()
        else:
            # polyfill with a No Data status
            result_status = GqlStatusObject._omitted_result()

        notification_status_objects = itertools.chain(
            notification_status_objects, (result_status,)
        )

        def status_precedence(status: GqlStatusObject) -> int:
            if status.gql_status.startswith("02"):
                # no data
                return 3
            if status.gql_status.startswith("01"):
                # warning
                return 2
            if status.gql_status.startswith("00"):
                # success
                return 1
            if status.gql_status.startswith("03"):
                # informational
                return 0
            return -1

        notification_status_objects = sorted(
            notification_status_objects,
            key=status_precedence,
            reverse=True,
        )
        self._gql_status_objects = tuple(notification_status_objects)

        return self._gql_status_objects


_COUNTER_KEY_TO_ATTR_NAME = {
    "nodes-created": "nodes_created",
    "nodes-deleted": "nodes_deleted",
    "relationships-created": "relationships_created",
    "relationships-deleted": "relationships_deleted",
    "properties-set": "properties_set",
    "labels-added": "labels_added",
    "labels-removed": "labels_removed",
    "indexes-added": "indexes_added",
    "indexes-removed": "indexes_removed",
    "constraints-added": "constraints_added",
    "constraints-removed": "constraints_removed",
    "system-updates": "system_updates",
    "contains-updates": "_contains_updates",
    "contains-system-updates": "_contains_system_updates",
}

_COUNTER_ATTR_NAME_TO_KEY = {
    v: k for k, v in _COUNTER_KEY_TO_ATTR_NAME.items()
}


class SummaryCounters:
    """Contains counters for various operations that a query triggered."""

    #:
    nodes_created: int = 0

    #:
    nodes_deleted: int = 0

    #:
    relationships_created: int = 0

    #:
    relationships_deleted: int = 0

    #:
    properties_set: int = 0

    #:
    labels_added: int = 0

    #:
    labels_removed: int = 0

    #:
    indexes_added: int = 0

    #:
    indexes_removed: int = 0

    #:
    constraints_added: int = 0

    #:
    constraints_removed: int = 0

    #:
    system_updates: int = 0

    _contains_updates = None
    _contains_system_updates = None

    def __init__(self, statistics) -> None:
        for key, value in dict(statistics).items():
            attr_name = _COUNTER_KEY_TO_ATTR_NAME.get(key)
            if attr_name:
                setattr(self, attr_name, value)

    def __repr__(self) -> str:
        statistics = {
            _COUNTER_ATTR_NAME_TO_KEY[k]: v
            for k, v in vars(self).items()
            if k in _COUNTER_ATTR_NAME_TO_KEY
        }
        return f"{self.__class__.__name__}({statistics!r})"

    def __str__(self) -> str:
        attrs = []
        for k, v in vars(self).items():
            if k.startswith("_"):  # hide private attributes
                continue
            if hasattr(self.__class__, k) and getattr(self.__class__, k) == v:
                # hide default values
                continue
            attrs.append(f"{k}: {v}")
        attrs.append(f"contains_updates: {self.contains_updates}")
        attrs.append(
            f"contains_system_updates: {self.contains_system_updates}"
        )
        return f"SummaryCounters{{{', '.join(attrs)}}}"

    @property
    def contains_updates(self) -> bool:
        """
        Check if any counters tracking graph updates are greater than 0.

        True if any of the counters except for system_updates, are greater
        than 0. Otherwise, False.
        """
        if self._contains_updates is not None:
            return self._contains_updates
        return bool(
            self.nodes_created
            or self.nodes_deleted
            or self.relationships_created
            or self.relationships_deleted
            or self.properties_set
            or self.labels_added
            or self.labels_removed
            or self.indexes_added
            or self.indexes_removed
            or self.constraints_added
            or self.constraints_removed
        )

    @property
    def contains_system_updates(self) -> bool:
        """True if the system database was updated, otherwise False."""
        if self._contains_system_updates is not None:
            return self._contains_system_updates
        return self.system_updates > 0


@dataclass
class SummaryInputPosition:
    """
    Structured form of a gql status/notification position.

    .. seealso::
        :attr:`.GqlStatusObject.position`,
        :attr:`.SummaryNotification.position`,
        :data:`.SummaryNotificationPosition`

    .. versionadded:: 5.22
    """

    #: The line number of the notification. Line numbers start at 1.
    line: int
    #: The column number of the notification. Column numbers start at 1.
    column: int
    #: The character offset of the notification. Offsets start at 0.
    offset: int

    @classmethod
    def _from_metadata(cls, metadata: object) -> te.Self | None:
        if not isinstance(metadata, dict):
            return None
        line = metadata.get("line")
        if not isinstance(line, int) or isinstance(line, bool):
            return None
        column = metadata.get("column")
        if not isinstance(column, int) or isinstance(column, bool):
            return None
        offset = metadata.get("offset")
        if not isinstance(offset, int) or isinstance(offset, bool):
            return None
        return cls(line=line, column=column, offset=offset)

    def __str__(self) -> str:
        return (
            f"line: {self.line}, column: {self.column}, offset: {self.offset}"
        )

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} "
            f"line={self.line!r}, "
            f"column={self.column!r}, "
            f"offset={self.offset!r}"
            ">"
        )


# Deprecated alias for :class:`.SummaryInputPosition`.
#
# .. versionadded:: 5.7
#
# .. versionchanged:: 5.22
#     Deprecated in favor of :class:`.SummaryInputPosition`.
SummaryNotificationPosition: te.TypeAlias = SummaryInputPosition


_SEVERITY_LOOKUP: dict[t.Any, NotificationSeverity] = {
    "WARNING": NotificationSeverity.WARNING,
    "INFORMATION": NotificationSeverity.INFORMATION,
}

_CATEGORY_LOOKUP: dict[t.Any, NotificationCategory] = {
    "HINT": NotificationCategory.HINT,
    "UNRECOGNIZED": NotificationCategory.UNRECOGNIZED,
    "UNSUPPORTED": NotificationCategory.UNSUPPORTED,
    "PERFORMANCE": NotificationCategory.PERFORMANCE,
    "DEPRECATION": NotificationCategory.DEPRECATION,
    "GENERIC": NotificationCategory.GENERIC,
    "SECURITY": NotificationCategory.SECURITY,
    "TOPOLOGY": NotificationCategory.TOPOLOGY,
    "SCHEMA": NotificationCategory.SCHEMA,
}

_CLASSIFICATION_LOOKUP: dict[t.Any, NotificationClassification] = {
    k: NotificationClassification(v) for k, v in _CATEGORY_LOOKUP.items()
}


if t.TYPE_CHECKING:

    class _SummaryNotificationKwargs(te.TypedDict, total=False):
        title: str
        code: str
        description: str
        severity_level: NotificationSeverity
        category: NotificationCategory
        raw_severity_level: str
        raw_category: str
        position: SummaryInputPosition | None


@dataclass
class SummaryNotification:
    """
    Structured form of a notification received from the server.

    .. seealso:: :attr:`.ResultSummary.summary_notifications`

    .. versionadded:: 5.7
    """

    title: str = ""
    code: str = ""
    description: str = ""
    severity_level: NotificationSeverity = NotificationSeverity.UNKNOWN
    category: NotificationCategory = NotificationCategory.UNKNOWN
    raw_severity_level: str = ""
    raw_category: str = ""
    position: SummaryNotificationPosition | None = None

    @classmethod
    def _from_metadata(cls, metadata: object) -> te.Self:
        if not isinstance(metadata, dict):
            return cls()
        kwargs: _SummaryNotificationKwargs = {
            "position": SummaryInputPosition._from_metadata(
                metadata.get("position")
            ),
        }
        str_keys: tuple[t.Literal["title", "code", "description"], ...] = (
            "title",
            "code",
            "description",
        )
        for key in str_keys:
            value = metadata.get(key)
            if isinstance(value, str):
                kwargs[key] = value
        severity = metadata.get("severity")
        if isinstance(severity, str):
            kwargs["raw_severity_level"] = severity
            kwargs["severity_level"] = _SEVERITY_LOOKUP.get(
                severity, NotificationSeverity.UNKNOWN
            )
        category = metadata.get("category")
        if isinstance(category, str):
            kwargs["raw_category"] = category
            kwargs["category"] = _CATEGORY_LOOKUP.get(
                category, NotificationCategory.UNKNOWN
            )
        return cls(**kwargs)

    def __str__(self) -> str:
        return (
            f"{{severity: {self.raw_severity_level}}} {{code: {self.code}}} "
            f"{{category: {self.raw_category}}} {{title: {self.title}}} "
            f"{{description: {self.description}}} "
            f"{{position: {self.position}}}"
        )


POLYFILL_DIAGNOSTIC_RECORD = (
    ("OPERATION", ""),
    ("OPERATION_CODE", "0"),
    ("CURRENT_SCHEMA", "/"),
)


_SUCCESS_STATUS_METADATA = {
    "gql_status": "00000",
    "status_description": "note: successful completion",
    "diagnostic_record": dict(POLYFILL_DIAGNOSTIC_RECORD),
}
_OMITTED_RESULT_STATUS_METADATA = {
    "gql_status": "00001",
    "status_description": "note: successful completion - omitted result",
    "diagnostic_record": dict(POLYFILL_DIAGNOSTIC_RECORD),
}
_NO_DATA_STATUS_METADATA = {
    "gql_status": "02000",
    "status_description": "note: no data",
    "diagnostic_record": dict(POLYFILL_DIAGNOSTIC_RECORD),
}


class GqlStatusObject:
    """
    Representation for GqlStatusObject found when executing a query.

    GqlStatusObjects are a superset of notifications, i.e., some but not all
    GqlStatusObjects are notifications.
    Notifications can be filtered server-side with
    driver config
    :ref:`driver-notifications-disabled-classifications-ref` and
    :ref:`driver-notifications-min-severity-ref` as well as
    session config
    :ref:`session-notifications-disabled-classifications-ref` and
    :ref:`session-notifications-min-severity-ref`.

    .. seealso:: :attr:`.ResultSummary.gql_status_objects`

    .. versionadded:: 5.22
    """

    # internal dictionaries, never handed to assure immutability
    _status_metadata: dict[str, t.Any]
    _status_diagnostic_record: dict[str, t.Any] | None = None

    _is_notification: bool
    _gql_status: str
    _status_description: str
    _position: SummaryInputPosition | None
    _raw_classification: str | None
    _classification: NotificationClassification
    _raw_severity: str | None
    _severity: NotificationSeverity
    _diagnostic_record: dict[str, t.Any]

    @classmethod
    def _success(cls) -> te.Self:
        obj = cls()
        obj._status_metadata = _SUCCESS_STATUS_METADATA
        return obj

    @classmethod
    def _omitted_result(cls) -> te.Self:
        obj = cls()
        obj._status_metadata = _OMITTED_RESULT_STATUS_METADATA
        return obj

    @classmethod
    def _no_data(cls) -> te.Self:
        obj = cls()
        obj._status_metadata = _NO_DATA_STATUS_METADATA
        return obj

    @classmethod
    def _from_status_metadata(cls, metadata: object) -> te.Self:
        obj = cls()
        if isinstance(metadata, dict):
            obj._status_metadata = metadata
        else:
            obj._status_metadata = {}
        return obj

    @classmethod
    def _from_notification_metadata(cls, metadata: object) -> te.Self:
        obj = cls()
        if not isinstance(metadata, dict):
            metadata = {}
        description = metadata.get("description")
        neo4j_code = metadata.get("neo4j_code")
        if not isinstance(neo4j_code, str):
            neo4j_code = ""
        title = metadata.get("title")
        if not isinstance(title, str):
            title = ""
        position = SummaryInputPosition._from_metadata(
            metadata.get("position")
        )
        classification = metadata.get("category")
        if not isinstance(classification, str):
            classification = None
        severity = metadata.get("severity")
        if not isinstance(severity, str):
            severity = None

        if severity == "WARNING":
            gql_status = "01N42"
            if not isinstance(description, str) or not description:
                description = "warn: unknown warning"
        else:
            # for "INFORMATION" or if severity is missing
            gql_status = "03N42"
            if not isinstance(description, str) or not description:
                description = "info: unknown notification"

        diagnostic_record = dict(POLYFILL_DIAGNOSTIC_RECORD)
        if "category" in metadata:
            diagnostic_record["_classification"] = metadata["category"]
        if "severity" in metadata:
            diagnostic_record["_severity"] = metadata["severity"]
        if "position" in metadata:
            diagnostic_record["_position"] = metadata["position"]

        obj._status_metadata = {
            "gql_status": gql_status,
            "status_description": description,
            "neo4j_code": neo4j_code,
            "title": title,
            "diagnostic_record": diagnostic_record,
        }
        obj._gql_status = gql_status
        obj._status_description = description
        obj._position = position
        obj._raw_classification = classification
        obj._raw_severity = severity
        obj._is_notification = True
        return obj

    def __str__(self) -> str:
        return self.status_description

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} "
            f"gql_status={self.gql_status!r}, "
            f"status_description={self.status_description!r}, "
            f"position={self.position!r}, "
            f"raw_classification={self.raw_classification!r}, "
            f"classification={self.classification!r}, "
            f"raw_severity={self.raw_severity!r}, "
            f"severity={self.severity!r}, "
            f"diagnostic_record={self.diagnostic_record!r}"
            ">"
        )

    @property
    def is_notification(self) -> bool:
        """
        Whether this GqlStatusObject is a notification.

        Only some GqlStatusObjects are notifications.
        The definition of notification is vendor-specific.
        Notifications are those GqlStatusObjects that provide additional
        information and can be filtered out via
        :ref:`driver-notifications-disabled-classifications-ref` and
        :ref:`driver-notifications-min-severity-ref` as well as.

        The fields :attr:`.position`,
        :attr:`.raw_classification`, :attr:`.classification`,
        :attr:`.raw_severity`, and :attr:`.severity` are only meaningful
        for notifications.
        """
        if hasattr(self, "_is_notification"):
            return self._is_notification

        neo4j_code = self._status_metadata.get("neo4j_code")
        self._is_notification = bool(
            isinstance(neo4j_code, str) and neo4j_code
        )
        return self._is_notification

    @classmethod
    def _extract_str_field(
        cls,
        data: dict[str, t.Any],
        key: str,
        default: _T = "",  # type: ignore[assignment]
    ) -> str | _T:
        value = data.get(key)
        if isinstance(value, str):
            return value
        else:
            return default

    @property
    def gql_status(self) -> str:
        """
        The GQLSTATUS.

        The following GQLSTATUS codes denote codes that the driver will use
        for polyfilling (when connected to an old, non-GQL-aware server).
        Further, they may be used by servers during the transition-phase to
        GQLSTATUS-awareness.

         * ``01N42`` (warning - unknown warning)
         * ``02N42`` (no data - unknown subcondition)
         * ``03N42`` (informational - unknown notification)
         * ``05N42`` (general processing exception - unknown error)

        .. note::
            This means these codes are not guaranteed to be stable and may
            change in future versions of the driver or the server.
        """
        if hasattr(self, "_gql_status"):
            return self._gql_status

        self._gql_status = self._extract_str_field(
            self._status_metadata, "gql_status"
        )
        return self._gql_status

    @property
    def status_description(self) -> str:
        """A description of the status."""
        if hasattr(self, "_status_description"):
            return self._status_description

        self._status_description = self._extract_str_field(
            self._status_metadata, "status_description"
        )
        return self._status_description

    def _get_status_diagnostic_record(self) -> dict[str, t.Any]:
        if self._status_diagnostic_record is not None:
            return self._status_diagnostic_record

        self._status_diagnostic_record = self._status_metadata.get(
            "diagnostic_record", {}
        )
        if not isinstance(self._status_diagnostic_record, dict):
            self._status_diagnostic_record = {}
        return self._status_diagnostic_record

    @property
    def position(self) -> SummaryInputPosition | None:
        """
        The position of the input that caused the status (if applicable).

        This is vendor-specific information.

        Only notifications (see :attr:`.is_notification`) have a meaningful
        position.

        The value is :data:`None` if the server's data was missing or could not
        be interpreted.
        """
        if hasattr(self, "_position"):
            return self._position

        diag_record = self._get_status_diagnostic_record()
        self._position = SummaryInputPosition._from_metadata(
            diag_record.get("_position")
        )
        return self._position

    @property
    def raw_classification(self) -> str | None:
        """
        The raw (``str``) classification of the status.

        This is a vendor-specific classification that can be used to filter
        notifications.

        Only notifications (see :attr:`.is_notification`) have a meaningful
        classification.
        """
        if hasattr(self, "_raw_classification"):
            return self._raw_classification

        diag_record = self._get_status_diagnostic_record()
        self._raw_classification = self._extract_str_field(
            diag_record, "_classification", None
        )
        return self._raw_classification

    @property
    def classification(self) -> NotificationClassification:
        """
        Parsed version of :attr:`.raw_classification`.

        Only notifications (see :attr:`.is_notification`) have a meaningful
        classification.
        """
        if hasattr(self, "_classification"):
            return self._classification

        self._classification = _CLASSIFICATION_LOOKUP.get(
            self.raw_classification, NotificationClassification.UNKNOWN
        )
        return self._classification

    @property
    def raw_severity(self) -> str | None:
        """
        The raw (``str``) severity of the status.

        This is a vendor-specific severity that can be used to filter
        notifications.

        Only notifications (see :attr:`.is_notification`) have a meaningful
        severity.
        """
        if hasattr(self, "_raw_severity"):
            return self._raw_severity

        diag_record = self._get_status_diagnostic_record()
        self._raw_severity = self._extract_str_field(
            diag_record, "_severity", None
        )
        return self._raw_severity

    @property
    def severity(self) -> NotificationSeverity:
        """
        Parsed version of :attr:`.raw_severity`.

        Only notifications (see :attr:`.is_notification`) have a meaningful
        severity.
        """
        if hasattr(self, "_severity"):
            return self._severity

        self._severity = _SEVERITY_LOOKUP.get(
            self.raw_severity, NotificationSeverity.UNKNOWN
        )
        return self._severity

    @property
    def diagnostic_record(self) -> dict[str, t.Any]:
        """Further information about the GQLSTATUS for diagnostic purposes."""
        if hasattr(self, "_diagnostic_record"):
            return self._diagnostic_record

        self._diagnostic_record = deepcopy(
            self._get_status_diagnostic_record()
        )
        return self._diagnostic_record



# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

from .. import _typing as t
from .._async_compat.concurrency import CooperativeLock
from .._async_compat.util import Util
from ..api import (
    BookmarkManager,
    Bookmarks,
)


TBmSupplier = t.Callable[[], Bookmarks | t.Union[Bookmarks]]
TBmConsumer = t.Callable[[Bookmarks], t.Union[None] | None]


class Neo4jBookmarkManager(BookmarkManager):
    def __init__(
        self,
        initial_bookmarks: Bookmarks | None = None,
        bookmarks_supplier: TBmSupplier | None = None,
        bookmarks_consumer: TBmConsumer | None = None,
    ) -> None:
        super().__init__()
        self._bookmarks_supplier = bookmarks_supplier
        self._bookmarks_consumer = bookmarks_consumer
        if not initial_bookmarks:
            self._bookmarks = set()
        else:
            self._bookmarks = set(initial_bookmarks.raw_values)
        self._lock = CooperativeLock()

    def update_bookmarks(
        self,
        previous_bookmarks: t.Collection[str],
        new_bookmarks: t.Collection[str],
    ) -> None:
        if not new_bookmarks:
            return
        with self._lock:
            self._bookmarks.difference_update(previous_bookmarks)
            self._bookmarks.update(new_bookmarks)
            if self._bookmarks_consumer:
                curr_bms_snapshot = Bookmarks.from_raw_values(self._bookmarks)
        if self._bookmarks_consumer:
            Util.callback(
                self._bookmarks_consumer, curr_bms_snapshot
            )

    def get_bookmarks(self) -> set[str]:
        with self._lock:
            bms = set(self._bookmarks)
        if self._bookmarks_supplier:
            extra_bms = Util.callback(self._bookmarks_supplier)
            bms.update(extra_bms.raw_values)
        return bms



# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

import asyncio
from functools import wraps
from inspect import isclass
from warnings import warn

from . import _typing as t
from .warnings import PreviewWarning


if t.TYPE_CHECKING:
    _FuncT = t.TypeVar("_FuncT", bound=t.Callable)


# Copy globals as function locals to make sure that they are available
# during Python shutdown when the Pool is destroyed.
def deprecation_warn(message, stack_level=1, _warn=warn):
    _warn(message, category=DeprecationWarning, stacklevel=stack_level + 1)


def deprecated(message: str) -> t.Callable[[_FuncT], _FuncT]:
    """
    Decorate deprecated functions and methods.

    ::

        @deprecated("'foo' has been deprecated in favour of 'bar'")
        def foo(x):
            pass

        @property
        @deprecated("'bar' will be internal without a replacement")
        def bar(self):
            return "bar"

        @property
        def baz(self):
            return self._baz

        @baz.setter
        @deprecated("'baz' will be read-only in the future")
        def baz(self, value):
            self._baz = value

    """
    return _make_warning_decorator(message, deprecation_warn)


def preview_warn(message, stack_level=1):
    message += (
        " It might be changed without following the deprecation policy. "
        "See also "
        "https://github.com/neo4j/neo4j-python-driver/wiki/preview-features."
    )
    warn(message, category=PreviewWarning, stacklevel=stack_level + 1)


def preview(message) -> t.Callable[[_FuncT], _FuncT]:
    """
    Decorate functions and methods as preview.

    ::

        @preview("foo is a preview.")
        def foo(x):
            pass
    """
    return _make_warning_decorator(message, preview_warn)


if t.TYPE_CHECKING:

    class _WarningFunc(t.Protocol):
        def __call__(self, message: str, stack_level: int = 1) -> None: ...
else:
    _WarningFunc = object


def _make_warning_decorator(
    message: str,
    warning_func: _WarningFunc,
) -> t.Callable[[_FuncT], _FuncT]:
    def decorator(f):
        if asyncio.iscoroutinefunction(f):

            @wraps(f)
            async def inner(*args, **kwargs):
                warning_func(message, stack_level=2)
                return await f(*args, **kwargs)

            inner._without_warning = f
            return inner
        if isclass(f):
            if hasattr(f, "__init__"):
                original_init = f.__init__

                @wraps(original_init)
                def inner(self, *args, **kwargs):
                    warning_func(message, stack_level=2)
                    return original_init(self, *args, **kwargs)

                def _without_warning(cls, *args, **kwargs):
                    obj = cls.__new__(cls, *args, **kwargs)
                    original_init(obj, *args, **kwargs)
                    return obj

                f.__init__ = inner
                f._without_warning = classmethod(_without_warning)
                return f
            raise TypeError("Cannot decorate class without __init__")
        else:

            @wraps(f)
            def inner(*args, **kwargs):
                warning_func(message, stack_level=2)
                return f(*args, **kwargs)

            inner._without_warning = f
            return inner

    return decorator


# Copy globals as function locals to make sure that they are available
# during Python shutdown when the Pool is destroyed.
def unclosed_resource_warn(obj, _warn=warn):
    cls_name = obj.__class__.__name__
    msg = f"unclosed  {cls_name}: {obj!r}."
    _warn(msg, ResourceWarning, stacklevel=2, source=obj)



# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections.abc import MutableSet
from contextlib import suppress
from logging import getLogger
from time import monotonic

from ._addressing import Address


log = getLogger("neo4j.pool")


class OrderedSet(MutableSet):
    def __init__(self, elements=()):
        # dicts keep insertion order starting with Python 3.7
        self._elements = dict.fromkeys(elements)
        self._current = None

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(("
            f"{', '.join(map(repr, self._elements))}"
            f"))"
        )

    def __str__(self):
        return f"{{{', '.join(map(repr, self._elements))}}}"

    def __contains__(self, element):
        return element in self._elements

    def __iter__(self):
        return iter(self._elements)

    def __len__(self):
        return len(self._elements)

    def __getitem__(self, index):
        return list(self._elements.keys())[index]

    def add(self, element):
        self._elements[element] = None

    def clear(self):
        self._elements.clear()

    def discard(self, element):
        with suppress(KeyError):
            del self._elements[element]

    def remove(self, element):
        try:
            del self._elements[element]
        except KeyError:
            raise ValueError(element) from None

    def update(self, elements=()):
        self._elements.update(dict.fromkeys(elements))

    def replace(self, elements=()):
        e = self._elements
        e.clear()
        e.update(dict.fromkeys(elements))


class RoutingTable:
    @classmethod
    def parse_routing_info(cls, *, database, servers, ttl):
        """
        Construct a new RoutingTable instance from the given routing info.

        The routing table info is returned from the procedure or BOLT message.
        """
        routers = []
        readers = []
        writers = []
        try:
            for server in servers:
                role = server["role"]
                addresses = [
                    Address.parse(address, default_port=7687)
                    for address in server["addresses"]
                ]
                if role == "ROUTE":
                    routers.extend(addresses)
                elif role == "READ":
                    readers.extend(addresses)
                elif role == "WRITE":
                    writers.extend(addresses)
        except (KeyError, TypeError) as exc:
            raise ValueError("Cannot parse routing info") from exc
        else:
            return cls(
                database=database,
                routers=routers,
                readers=readers,
                writers=writers,
                ttl=ttl,
            )

    def __init__(self, *, database, routers=(), readers=(), writers=(), ttl=0):
        self.initial_routers = OrderedSet(routers)
        self.routers = OrderedSet(routers)
        self.readers = OrderedSet(readers)
        self.writers = OrderedSet(writers)
        self.initialized_without_writers = not self.writers
        self.last_updated_time = monotonic()
        self.ttl = ttl
        self.database = database

    def __repr__(self):
        return (
            "RoutingTable("
            f"database={self.database!r}, "
            f"routers={tuple(self.routers)!r}, "
            f"readers={tuple(self.readers)!r}, "
            f"writers={tuple(self.writers)!r}, "
            f"last_updated_time={self.last_updated_time!r}, "
            f"ttl={self.ttl!r}"
            ")"
        )

    def __contains__(self, address):
        return (
            address in self.routers
            or address in self.readers
            or address in self.writers
        )

    def is_fresh(self, readonly=False):
        """Indicate whether routing information is still usable."""
        assert isinstance(readonly, bool)
        expired = self.last_updated_time + self.ttl <= monotonic()
        if readonly:
            has_server_for_mode = bool(self.readers)
        else:
            has_server_for_mode = bool(self.writers)
        res = not expired and self.routers and has_server_for_mode
        log.debug(
            "[#0000]  _: <ROUTING> checking table freshness "
            "(readonly=%r): table expired=%r, "
            "has_server_for_mode=%r, table routers=%r => %r",
            readonly,
            expired,
            has_server_for_mode,
            self.routers,
            res,
        )
        return res

    def should_be_purged_from_memory(self):
        """
        Check if routing table needs to be purged from memory.

        This is the case if the routing table is stale and has not been used
        for a long time.

        :returns: Returns true if it is old and not used for a while.
        :rtype: bool
        """
        from ._conf import RoutingConfig

        perf_time = monotonic()
        valid_until = (
            self.last_updated_time
            + self.ttl
            + RoutingConfig.routing_table_purge_delay
        )
        should_be_purged = valid_until <= perf_time
        log.debug(
            "[#0000]  _: <ROUTING> purge check: "
            "last_updated_time=%r, ttl=%r, perf_time=%r => %r",
            self.last_updated_time,
            self.ttl,
            perf_time,
            should_be_purged,
        )
        return should_be_purged

    def update(self, new_routing_table):
        """Update the routing table with new routing information."""
        self.routers.replace(new_routing_table.routers)
        self.readers.replace(new_routing_table.readers)
        self.writers.replace(new_routing_table.writers)
        self.initialized_without_writers = not self.writers
        self.last_updated_time = monotonic()
        self.ttl = new_routing_table.ttl
        log.debug("[#0000]  _: <ROUTING> updated table=%r", self)

    def servers(self):
        return set(self.routers) | set(self.writers) | set(self.readers)

    def __eq__(self, other):
        if not isinstance(other, RoutingTable):
            return NotImplemented
        return (
            self.database == other.database
            and self.routers == other.routers
            and self.readers == other.readers
            and self.writers == other.writers
            and self.ttl == other.ttl
        )



# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ....vector import (
    Vector,
    VectorDType,
)
from ...packstream import Structure


_DTYPE_LOOKUP = {
    b"\xc8": VectorDType.I8,
    b"\xc9": VectorDType.I16,
    b"\xca": VectorDType.I32,
    b"\xcb": VectorDType.I64,
    b"\xc6": VectorDType.F32,
    b"\xc1": VectorDType.F64,
}

_TYPE_LOOKUP = {v: k for k, v in _DTYPE_LOOKUP.items()}


def hydrate_vector(typ: bytes, data: bytes) -> Vector:
    """
    Hydrator for `Vector` values.

    :param typ: bytes marking the inner type of the vector
    :param data: big-endian bytes representing the vector data
    :returns: Vector
    """
    dtype = _DTYPE_LOOKUP[typ]
    return Vector(data, dtype)


def dehydrate_vector(value: Vector) -> Structure:
    """
    Dehydrator for `Vector` values.

    :param value:
    :type value: Vector
    :returns:
    """
    return Structure(b"V", _TYPE_LOOKUP[value.dtype], value.raw())



# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

from . import _typing as _t
from ._debug import NotificationPrinter


if _t.TYPE_CHECKING:
    from ._work.summary import GqlStatusObject


__all__ = [
    "Neo4jDeprecationWarning",
    "Neo4jWarning",
    "PreviewWarning",
]


class PreviewWarning(Warning):
    """
    A driver feature in preview has been used.

    It might be changed without following the deprecation policy.
    See also https://github.com/neo4j/neo4j-python-driver/wiki/preview-features

    .. versionadded:: 5.8

    .. versionchanged:: 6.0
        Moved from ``neo4j.PreviewWarning`` to
        ``neo4j.warnings.PreviewWarning``.
    """


class Neo4jWarning(Warning):
    """
    Warning emitted for notifications sent by the server.

    Which notifications trigger a warning can be controlled by a
    configuration option: :ref:`driver-warn-notification-severity-ref`

    **This is a preview**.
    It might be changed without following the deprecation policy.

    See also
    https://github.com/neo4j/neo4j-python-driver/wiki/preview-features

    :param notification: The notification that triggered the warning.
    :param query: The query for which the notification was sent.
        If provided, it will be used for a more detailed warning message.

    .. versionadded:: 5.21

    .. versionchanged:: 6.0
        :attr:`.notification` is now of type :class:`.SummaryNotification`.

    .. seealso:: :ref:`development-environment-ref`
    """

    #: The notification that triggered the warning.
    notification: GqlStatusObject

    def __init__(
        self,
        notification: GqlStatusObject,
        query: str | None = None,
    ) -> None:
        msg = str(NotificationPrinter(notification, query))
        super().__init__(msg)
        self.notification = notification


class Neo4jDeprecationWarning(Neo4jWarning, DeprecationWarning):
    """
    Warning emitted for deprecation notifications sent by the server.

    .. note::

        This warning is a subclass of :class:`DeprecationWarning`.
        This means that Python will not show this warning by default.

    **This is a preview**.
    It might be changed without following the deprecation policy.

    See also
    https://github.com/neo4j/neo4j-python-driver/wiki/preview-features

    :param notification: The notification that triggered the warning.

    .. versionadded:: 5.21
    """



# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ruff: noqa: N818
# Not going to rename all Error classes that don't end on Error,
# which would break pretty much all users just to please the linter.


"""
Module containing the core driver exceptions.

Driver API Errors
=================
+ Neo4jError
  + ClientError
    + CypherSyntaxError
    + CypherTypeError
    + ConstraintError
    + AuthError
      + TokenExpired
    + Forbidden
  + DatabaseError
  + TransientError
    + DatabaseUnavailable
    + NotALeader
    + ForbiddenOnReadOnlyDatabase

+ DriverError
  + SessionError
  + TransactionError
    + TransactionNestingError
  + ResultError
    + ResultFailedError
    + ResultConsumedError
    + ResultNotSingleError
  + BrokenRecordError
  + SessionExpired
  + ServiceUnavailable
    + RoutingServiceUnavailable
    + WriteServiceUnavailable
    + ReadServiceUnavailable
    + IncompleteCommit
  + ConfigurationError
    + AuthConfigurationError
    + CertificateConfigurationError
    + UnsupportedServerProduct
  + ConnectionPoolError
    + ConnectionAcquisitionTimeoutError
"""

from __future__ import annotations as _

from copy import deepcopy as _deepcopy
from enum import Enum as _Enum

from . import _typing as _t


if _t.TYPE_CHECKING:
    from ._async.work import (
        AsyncManagedTransaction as _AsyncManagedTransaction,
        AsyncResult as _AsyncResult,
        AsyncSession as _AsyncSession,
        AsyncTransaction as _AsyncTransaction,
    )
    from ._sync.work import (
        ManagedTransaction as _ManagedTransaction,
        Result as _Result,
        Session as _Session,
        Transaction as _Transaction,
    )

    _TTransaction: _t.TypeAlias = (
        _AsyncManagedTransaction
        | _AsyncTransaction
        | _ManagedTransaction
        | _Transaction
    )
    _TResult: _t.TypeAlias = _AsyncResult | _Result
    _TSession: _t.TypeAlias = _AsyncSession | _Session
    _T = _t.TypeVar("_T")


__all__ = [
    "AuthConfigurationError",
    "AuthError",
    "BrokenRecordError",
    "CertificateConfigurationError",
    "ClientError",
    "ConfigurationError",
    "ConnectionAcquisitionTimeoutError",
    "ConnectionPoolError",
    "ConstraintError",
    "CypherSyntaxError",
    "CypherTypeError",
    "DatabaseError",
    "DatabaseUnavailable",
    "DriverError",
    "Forbidden",
    "ForbiddenOnReadOnlyDatabase",
    "GqlError",
    "GqlErrorClassification",
    "IncompleteCommit",
    "Neo4jError",
    "NotALeader",
    "ReadServiceUnavailable",
    "ResultConsumedError",
    "ResultError",
    "ResultFailedError",
    "ResultNotSingleError",
    "RoutingServiceUnavailable",
    "ServiceUnavailable",
    "SessionError",
    "SessionExpired",
    "TokenExpired",
    "TransactionError",
    "TransactionNestingError",
    "TransientError",
    "UnsupportedServerProduct",
    "WriteServiceUnavailable",
]


_CLASSIFICATION_CLIENT: _t.Final[str] = "ClientError"
_CLASSIFICATION_TRANSIENT: _t.Final[str] = "TransientError"
_CLASSIFICATION_DATABASE: _t.Final[str] = "DatabaseError"


_ERROR_REWRITE_MAP: dict[str, tuple[str, str | None]] = {
    # This error can be retried ed. The driver just needs to re-authenticate
    # with the same credentials.
    "Neo.ClientError.Security.AuthorizationExpired": (
        _CLASSIFICATION_TRANSIENT,
        None,
    ),
    # In 5.0, this error has been re-classified as ClientError.
    # For backwards compatibility with Neo4j 4.4 and earlier, we re-map it in
    # the driver, too.
    "Neo.TransientError.Transaction.Terminated": (
        _CLASSIFICATION_CLIENT,
        "Neo.ClientError.Transaction.Terminated",
    ),
    # In 5.0, this error has been re-classified as ClientError.
    # For backwards compatibility with Neo4j 4.4 and earlier, we re-map it in
    # the driver, too.
    "Neo.TransientError.Transaction.LockClientStopped": (
        _CLASSIFICATION_CLIENT,
        "Neo.ClientError.Transaction.LockClientStopped",
    ),
}


_UNKNOWN_NEO4J_CODE: _t.Final[str] = "Neo.DatabaseError.General.UnknownError"
# TODO: 7.0 - Make _UNKNOWN_GQL_MESSAGE the default message
_UNKNOWN_MESSAGE: _t.Final[str] = "An unknown error occurred"
_UNKNOWN_GQL_STATUS: _t.Final[str] = "50N42"
_UNKNOWN_GQL_DESCRIPTION: _t.Final[str] = (
    "error: general processing exception - unexpected error"
)
_UNKNOWN_GQL_MESSAGE: _t.Final[str] = (
    f"{_UNKNOWN_GQL_STATUS}: "
    "Unexpected error has occurred. See debug log for details."
)
_UNKNOWN_GQL_DIAGNOSTIC_RECORD: _t.Final[tuple[tuple[str, _t.Any], ...]] = (
    ("OPERATION", ""),
    ("OPERATION_CODE", "0"),
    ("CURRENT_SCHEMA", "/"),
)


class GqlErrorClassification(str, _Enum):
    """
    Server-side GQL error category.

    Inherits from :class:`str` and :class:`enum.Enum`.
    Hence, can also be compared to its string value::

        >>> GqlErrorClassification.CLIENT_ERROR == "CLIENT_ERROR"
        True
        >>> GqlErrorClassification.DATABASE_ERROR == "DATABASE_ERROR"
        True
        >>> GqlErrorClassification.TRANSIENT_ERROR == "TRANSIENT_ERROR"
        True

    .. seealso:: :attr:`.GqlError.gql_classification`

    .. versionadded:: 5.26

    .. versionchanged:: 6.0 Stabilized from preview.
    """

    CLIENT_ERROR = "CLIENT_ERROR"
    DATABASE_ERROR = "DATABASE_ERROR"
    TRANSIENT_ERROR = "TRANSIENT_ERROR"
    #: Used when the server provides a Classification which the driver is
    #: unaware of.
    #: This can happen when connecting to a server newer than the driver or
    #: before GQL errors were introduced.
    UNKNOWN = "UNKNOWN"


class GqlError(Exception):
    """
    The GQL compliant data of an error.

    This error isn't raised by the driver as is.
    Instead, only subclasses are raised.
    Further, it is used as the :attr:`__cause__` of GqlError subclasses.

    Sometimes it is helpful or necessary to traverse the cause chain of
    GqlErrors to fully understand or appropriately handle the error. In such
    cases, users can either traverse the :attr:`__cause__` attribute of the
    error(s) or use the helper method :meth:`.find_by_gql_status`. Note that
    :attr:`__cause__` is a standard attribute of all Python
    :class:`BaseException` s: the cause chain may contain other exception types
    besides GqlError.

    .. versionadded: 5.26

    .. versionchanged:: 6.0 Stabilized from preview.
    """

    _gql_status: str
    _message: str
    _gql_status_description: str
    _gql_raw_classification: str | None
    _gql_classification: GqlErrorClassification
    _status_diagnostic_record: dict[str, _t.Any]  # original, internal only
    _diagnostic_record: dict[str, _t.Any]  # copy to be used externally
    _gql_cause: GqlError | None

    __cause__: BaseException | None
    """
    The GqlError's cause, if any.

    Sometimes it is helpful or necessary to traverse the cause chain of
    GqlErrors to fully understand or appropriately handle the error.

    .. seealso:: :meth:`.find_by_gql_status`
    """

    @staticmethod
    def _hydrate_cause(**metadata: _t.Any) -> GqlError:
        meta_extractor = _MetaExtractor(metadata)
        gql_status = meta_extractor.str_value("gql_status")
        description = meta_extractor.str_value("description")
        message = meta_extractor.str_value("message")
        if gql_status is None or description is None or message is None:
            gql_status = _UNKNOWN_GQL_STATUS
            message = _UNKNOWN_GQL_MESSAGE
            description = _UNKNOWN_GQL_DESCRIPTION
        diagnostic_record = meta_extractor.map_value("diagnostic_record")
        cause_map = meta_extractor.map_value("cause")
        if cause_map is not None:
            cause = GqlError._hydrate_cause(**cause_map)
        else:
            cause = None
        inst = GqlError()
        inst._init_gql(
            gql_status=gql_status,
            message=message,
            description=description,
            diagnostic_record=diagnostic_record,
            cause=cause,
        )
        return inst

    def _init_gql(
        self,
        *,
        gql_status: str,
        message: str,
        description: str,
        diagnostic_record: dict[str, _t.Any] | None = None,
        cause: GqlError | None = None,
    ) -> None:
        self._gql_status = gql_status
        self._message = message
        self._gql_status_description = description
        if diagnostic_record is not None:
            self._status_diagnostic_record = diagnostic_record
        self._gql_cause = cause

    def _set_gql_unknown(self) -> None:
        self._gql_status = _UNKNOWN_GQL_STATUS
        self._message = _UNKNOWN_GQL_MESSAGE
        self._gql_status_description = _UNKNOWN_GQL_DESCRIPTION

    def __getattribute__(self, item):
        if item != "__cause__":
            return super().__getattribute__(item)
        gql_cause = self._get_attr_or_none("_gql_cause")
        if gql_cause is None:
            # No GQL cause, no magic needed
            return super().__getattribute__(item)
        local_cause = self._get_attr_or_none("__cause__")
        if local_cause is None:
            # We have a GQL cause but no local cause
            # => set the GQL cause as the local cause
            self.__cause__ = gql_cause
            self.__suppress_context__ = True
            self._gql_cause = None
            return super().__getattribute__(item)
        # We have both a GQL cause and a local cause
        # => traverse the cause chain and append the local cause.
        root = gql_cause
        seen_errors = {id(self), id(root)}
        while True:
            cause = getattr(root, "__cause__", None)
            if cause is None:
                root.__cause__ = local_cause
                root.__suppress_context__ = True
                self.__cause__ = gql_cause
                self.__suppress_context__ = True
                self._gql_cause = None
                return gql_cause
            root = cause
            if id(root) in seen_errors:
                # Circular cause chain -> we have no choice but to either
                # overwrite the cause or ignore the new one.
                return local_cause
            seen_errors.add(id(root))

    def _get_attr_or_none(self, item):
        try:
            return super().__getattribute__(item)
        except AttributeError:
            return None

    @property
    def gql_status(self) -> str:
        """
        The GQLSTATUS returned from the server.

        The status code ``50N42`` (unknown error) is a special code that the
        driver will use for polyfilling (when connected to an old,
        non-GQL-aware server).
        Further, it may be used by servers during the transition-phase to
        GQLSTATUS-awareness.

        .. note::
            This means that the code ``50N42`` is not guaranteed to be stable
            and may change in future versions of the driver or the server.
        """
        if hasattr(self, "_gql_status"):
            return self._gql_status
        self._set_gql_unknown()
        return self._gql_status

    @property
    def message(self) -> str:
        """
        The error message returned by the server.

        It is a string representation of the error that occurred.

        This message is meant for human consumption and debugging purposes.
        Don't rely on it in a programmatic way.

        This value is never :data:`None` unless the subclass in question
        states otherwise.
        """
        if hasattr(self, "_message"):
            return self._message
        self._set_gql_unknown()
        return self._message

    @property
    def gql_status_description(self) -> str:
        """
        A description of the GQLSTATUS returned from the server.

        It describes the error that occurred in detail.

        This description is meant for human consumption and debugging purposes.
        Don't rely on it in a programmatic way.
        """
        if hasattr(self, "_gql_status_description"):
            return self._gql_status_description
        self._set_gql_unknown()
        return self._gql_status_description

    @property
    def gql_raw_classification(self) -> str | None:
        """
        Vendor specific classification of the error.

        This is a convenience accessor for ``_classification`` in the
        diagnostic record. :data:`None` is returned if the classification is
        not available or not a string.
        """
        if hasattr(self, "_gql_raw_classification"):
            return self._gql_raw_classification

        diag_record = self._get_status_diagnostic_record()
        classification = diag_record.get("_classification")
        if not isinstance(classification, str):
            self._gql_raw_classification = None
        else:
            self._gql_raw_classification = classification
        return self._gql_raw_classification

    @property
    def gql_classification(self) -> GqlErrorClassification:
        """The stable GqlErrorClassification for this error."""
        if hasattr(self, "_gql_classification"):
            return self._gql_classification
        classification = self.gql_raw_classification
        if not (
            isinstance(classification, str)
            and classification
            in _t.cast(_t.Iterable[str], iter(GqlErrorClassification))
        ):
            self._gql_classification = GqlErrorClassification.UNKNOWN
        else:
            self._gql_classification = GqlErrorClassification(classification)
        return self._gql_classification

    @property
    def diagnostic_record(self) -> _t.Mapping[str, _t.Any]:
        """The diagnostic record for this error."""
        if hasattr(self, "_diagnostic_record"):
            return self._diagnostic_record
        self._diagnostic_record = _deepcopy(
            self._get_status_diagnostic_record()
        )
        return self._diagnostic_record

    def _get_status_diagnostic_record(self) -> dict[str, _t.Any]:
        if hasattr(self, "_status_diagnostic_record"):
            return self._status_diagnostic_record

        self._status_diagnostic_record = dict(_UNKNOWN_GQL_DIAGNOSTIC_RECORD)
        return self._status_diagnostic_record

    def find_by_gql_status(self, status: str) -> GqlError | None:
        """
        Return the first GqlError in the cause chain with the given GQL status.

        This method traverses this GQLErorrs's :attr:`__cause__` chain,
        starting with this error itself, and returns the first error that has
        the given GQL status. If no error matches, :data:`None` is returned.

        Example::

            def invalid_syntax(err: GqlError) -> bool:
                return err.find_by_gql_status("42001") is not None

        :param status: The GQL status to search for.

        :returns: The first matching error or :data:`None`.

        .. versionadded:: 6.0
        """
        if self.gql_status == status:
            return self

        cause = self.__cause__
        while cause is not None:
            if isinstance(cause, GqlError) and cause.gql_status == status:
                return cause
            cause = getattr(cause, "__cause__", None)

        return None

    def __str__(self):
        return (
            f"{{gql_status: {self.gql_status}}} "
            f"{{gql_status_description: {self.gql_status_description}}} "
            f"{{message: {self.message}}} "
            f"{{diagnostic_record: {self.diagnostic_record}}} "
            f"{{raw_classification: {self.gql_raw_classification}}}"
        )


# Neo4jError
class Neo4jError(GqlError):
    """Raised when the Cypher engine returns an error to the client."""

    _neo4j_code: str
    _classification: str
    _category: str
    _title: str
    #: (dict) Any additional information returned by the server.
    _metadata: dict[str, _t.Any]
    _from_server: bool

    _retryable = False

    def __init__(self, *args: object) -> None:
        Exception.__init__(self, *args)
        self._from_server = False
        self._neo4j_code = _UNKNOWN_NEO4J_CODE
        self._message = _UNKNOWN_MESSAGE
        _, self._classification, self._category, self._title = (
            self._neo4j_code.split(".")
        )

    @staticmethod
    def _hydrate_neo4j(**metadata: _t.Any) -> Neo4jError:
        meta_extractor = _MetaExtractor(metadata)
        code = meta_extractor.str_value("code") or _UNKNOWN_NEO4J_CODE
        message = meta_extractor.str_value("message") or _UNKNOWN_MESSAGE
        inst = Neo4jError._basic_hydrate(
            neo4j_code=code,
            message=message,
        )
        inst._init_gql(
            gql_status=_UNKNOWN_GQL_STATUS,
            message=message,
            description=f"{_UNKNOWN_GQL_DESCRIPTION}. {message}",
        )
        inst._metadata = meta_extractor.rest()
        return inst

    @staticmethod
    def _hydrate_gql(**metadata: _t.Any) -> Neo4jError:
        meta_extractor = _MetaExtractor(metadata)
        gql_status = meta_extractor.str_value("gql_status")
        status_description = meta_extractor.str_value("description")
        message = meta_extractor.str_value("message")
        if gql_status is None or status_description is None or message is None:
            gql_status = _UNKNOWN_GQL_STATUS
            # TODO: 7.0 - Make this fall back to _UNKNOWN_GQL_MESSAGE
            message = _UNKNOWN_MESSAGE
            status_description = _UNKNOWN_GQL_DESCRIPTION
        neo4j_code = meta_extractor.str_value(
            "neo4j_code",
            _UNKNOWN_NEO4J_CODE,
        )
        diagnostic_record = meta_extractor.map_value("diagnostic_record")
        cause_map = meta_extractor.map_value("cause")
        if cause_map is not None:
            cause = Neo4jError._hydrate_cause(**cause_map)
        else:
            cause = None

        inst = Neo4jError._basic_hydrate(
            neo4j_code=neo4j_code,
            message=message,
        )
        inst._init_gql(
            gql_status=gql_status,
            message=message,
            description=status_description,
            diagnostic_record=diagnostic_record,
            cause=cause,
        )
        inst._metadata = meta_extractor.rest()

        return inst

    @staticmethod
    def _basic_hydrate(*, neo4j_code: str, message: str) -> Neo4jError:
        try:
            _, classification, category, title = neo4j_code.split(".")
        except ValueError:
            classification = _CLASSIFICATION_DATABASE
            category = "General"
            title = "UnknownError"
        else:
            classification_override, code_override = _ERROR_REWRITE_MAP.get(
                neo4j_code, (None, None)
            )
            if classification_override is not None:
                classification = classification_override
            if code_override is not None:
                neo4j_code = code_override

        error_class: type[Neo4jError] = Neo4jError._extract_error_class(
            classification, neo4j_code
        )

        assert issubclass(error_class, Exception)
        inst = Exception.__new__(error_class)
        inst._from_server = True
        inst._neo4j_code = neo4j_code
        inst._classification = classification
        inst._category = category
        inst._title = title
        inst._message = message

        return inst

    @staticmethod
    def _extract_error_class(classification, code) -> type[Neo4jError]:
        if classification == _CLASSIFICATION_CLIENT:
            try:
                return _client_errors[code]
            except KeyError:
                return ClientError

        elif classification == _CLASSIFICATION_TRANSIENT:
            try:
                return _transient_errors[code]
            except KeyError:
                return TransientError

        elif classification == _CLASSIFICATION_DATABASE:
            return DatabaseError

        else:
            return Neo4jError

    @property
    def message(self) -> str:
        """The error message returned by the server."""
        return self._message

    @property
    def code(self) -> str:
        """
        The neo4j error code returned by the server.

        For example, "Neo.ClientError.Security.AuthorizationExpired".
        """
        return self._neo4j_code

    @property
    def classification(self) -> str:
        # Undocumented, will likely be removed with support for neo4j codes
        return self._classification

    @property
    def category(self) -> str:
        # Undocumented, will likely be removed with support for neo4j codes
        return self._category

    @property
    def title(self) -> str:
        # Undocumented, will likely be removed with support for neo4j codes
        return self._title

    @property
    def metadata(self) -> dict[str, _t.Any]:
        # Undocumented, might be useful for debugging
        return self._metadata

    def is_retryable(self) -> bool:
        """
        Whether the error is retryable.

        Indicates whether a transaction that yielded this error makes sense to
        retry. This method makes mostly sense when implementing a custom
        retry policy in conjunction with :ref:`explicit-transactions-ref`.

        .. warning::

            Auto-commit transactions
            (:meth:`.Session.run`/:meth:`.AsyncSession.run`) are not retryable
            regardless of this value.

        :returns: :data:`True` if the error is retryable,
            :data:`False` otherwise.

        .. versionadded:: 5.0
        """
        return self._retryable

    def _unauthenticates_all_connections(self) -> bool:
        return (
            self._neo4j_code == "Neo.ClientError.Security.AuthorizationExpired"
        )

    def _is_fatal_during_discovery(self) -> bool:
        # checks if the code is an error that is caused by the client. In this
        # case the driver should fail fast during discovery.
        code = self._neo4j_code
        if not isinstance(code, str):
            return False
        if code in {
            "Neo.ClientError.Database.DatabaseNotFound",
            "Neo.ClientError.Transaction.InvalidBookmark",
            "Neo.ClientError.Transaction.InvalidBookmarkMixture",
            "Neo.ClientError.Statement.TypeError",
            "Neo.ClientError.Statement.ArgumentError",
            "Neo.ClientError.Request.Invalid",
        }:
            return True
        return (
            code.startswith("Neo.ClientError.Security.")
            and code != "Neo.ClientError.Security.AuthorizationExpired"
        )

    def _has_security_code(self) -> bool:
        if self._neo4j_code is None:
            return False
        return self._neo4j_code.startswith("Neo.ClientError.Security.")

    def __str__(self):
        if not getattr(self, "_from_server", False):
            return Exception.__str__(self)
        code = self._neo4j_code
        message = self._message
        # TODO: 7.0 - Check if including neo4j_code is still useful
        gql_status = self._gql_status
        gql_description = self._gql_status_description
        return (
            f"{{neo4j_code: {code}}} "
            f"{{message: {message}}} "
            f"{{gql_status: {gql_status}}} "
            f"{{gql_status_description: {gql_description}}}"
        )


class _MetaExtractor:
    def __init__(self, metadata: dict[str, _t.Any]) -> None:
        self._metadata = metadata

    def rest(self) -> dict[str, _t.Any]:
        return self._metadata

    @_t.overload
    def str_value(self, key: str) -> str | None: ...

    @_t.overload
    def str_value(self, key: str, default: _T) -> str | _T: ...

    def str_value(
        self, key: str, default: _T | None = None
    ) -> str | _T | None:
        res = self._metadata.pop(key, default)
        if not isinstance(res, str):
            res = default
        return res

    @_t.overload
    def map_value(self, key: str) -> dict[str, _t.Any] | None: ...

    @_t.overload
    def map_value(self, key: str, default: _T) -> dict[str, _t.Any] | _T: ...

    def map_value(
        self, key: str, default: _T | None = None
    ) -> dict[str, _t.Any] | _T | None:
        res = self._metadata.pop(key, default)
        if not (
            isinstance(res, dict) and all(isinstance(k, str) for k in res)
        ):
            res = default
        return res


# Neo4jError > ClientError
class ClientError(Neo4jError):
    """
    Bad client request.

    The Client sent a bad request - changing the request might yield a
    successful outcome.
    """


# Neo4jError > ClientError > CypherSyntaxError
class CypherSyntaxError(ClientError):
    pass


# Neo4jError > ClientError > CypherTypeError
class CypherTypeError(ClientError):
    pass


# Neo4jError > ClientError > ConstraintError
class ConstraintError(ClientError):
    pass


# Neo4jError > ClientError > AuthError
class AuthError(ClientError):
    """Raised when authentication failure occurs."""


# Neo4jError > ClientError > AuthError > TokenExpired
class TokenExpired(AuthError):
    """Raised when the authentication token has expired."""


# Neo4jError > ClientError > Forbidden
class Forbidden(ClientError):
    pass


# Neo4jError > DatabaseError
class DatabaseError(Neo4jError):
    """The database failed to service the request."""


# Neo4jError > TransientError
class TransientError(Neo4jError):
    """
    Transient Error.

    The database cannot service the request right now, retrying later might
    yield a successful outcome.
    """

    _retryable = True


# Neo4jError > TransientError > DatabaseUnavailable
class DatabaseUnavailable(TransientError):
    pass


# Neo4jError > TransientError > NotALeader
class NotALeader(TransientError):
    pass


# Neo4jError > TransientError > ForbiddenOnReadOnlyDatabase
class ForbiddenOnReadOnlyDatabase(TransientError):
    pass


_client_errors: dict[str, type[Neo4jError]] = {
    # ConstraintError
    "Neo.ClientError.Schema.ConstraintValidationFailed": ConstraintError,
    "Neo.ClientError.Schema.ConstraintViolation": ConstraintError,
    "Neo.ClientError.Statement.ConstraintVerificationFailed": ConstraintError,
    "Neo.ClientError.Statement.ConstraintViolation": ConstraintError,
    # CypherSyntaxError
    "Neo.ClientError.Statement.InvalidSyntax": CypherSyntaxError,
    "Neo.ClientError.Statement.SyntaxError": CypherSyntaxError,
    # CypherTypeError
    "Neo.ClientError.Procedure.TypeError": CypherTypeError,
    "Neo.ClientError.Statement.InvalidType": CypherTypeError,
    "Neo.ClientError.Statement.TypeError": CypherTypeError,
    # Forbidden
    "Neo.ClientError.General.ForbiddenOnReadOnlyDatabase": ForbiddenOnReadOnlyDatabase,  # noqa: E501
    "Neo.ClientError.General.ReadOnly": Forbidden,
    "Neo.ClientError.Schema.ForbiddenOnConstraintIndex": Forbidden,
    "Neo.ClientError.Schema.IndexBelongsToConstraint": Forbidden,
    "Neo.ClientError.Security.Forbidden": Forbidden,
    "Neo.ClientError.Transaction.ForbiddenDueToTransactionType": Forbidden,
    # AuthError
    "Neo.ClientError.Security.AuthorizationFailed": AuthError,
    "Neo.ClientError.Security.Unauthorized": AuthError,
    # TokenExpired
    "Neo.ClientError.Security.TokenExpired": TokenExpired,
    # NotALeader
    "Neo.ClientError.Cluster.NotALeader": NotALeader,
}

_transient_errors: dict[str, type[Neo4jError]] = {
    # DatabaseUnavailableError
    "Neo.TransientError.General.DatabaseUnavailable": DatabaseUnavailable
}


# DriverError
class DriverError(GqlError):
    """Raised when the Driver raises an error."""

    def is_retryable(self) -> bool:
        """
        Whether the error is retryable.

        Indicates whether a transaction that yielded this error makes sense to
        retry. This method makes mostly sense when implementing a custom
        retry policy in conjunction with :ref:`explicit-transactions-ref`.

        .. warning::

            Auto-commit transactions
            (:meth:`.Session.run`/:meth:`.AsyncSession.run`) are not retryable
            regardless of this value.

        :returns: :data:`True` if the error is retryable,
            :data:`False` otherwise.

        .. versionadded:: 5.0
        """
        return False

    def __str__(self):
        return Exception.__str__(self)


# DriverError > SessionError
class SessionError(DriverError):
    """Raised when an error occurs while using a session."""

    session: _TSession

    def __init__(self, session_, *args: object) -> None:
        super().__init__(*args)
        self.session = session_


# DriverError > TransactionError
class TransactionError(DriverError):
    """Raised when an error occurs while using a transaction."""

    transaction: _TTransaction

    def __init__(self, transaction_, *args: object) -> None:
        super().__init__(*args)
        self.transaction = transaction_


# DriverError > TransactionError > TransactionNestingError
class TransactionNestingError(TransactionError):
    """Raised when transactions are nested incorrectly."""


# DriverError > ResultError
class ResultError(DriverError):
    """Raised when an error occurs while using a result object."""

    result: _TResult

    def __init__(self, result_, *args: object) -> None:
        super().__init__(*args)
        self.result = result_


# DriverError > ResultError > ResultFailedError
class ResultFailedError(ResultError):
    """
    Raised when trying to access records of a failed result.

    A :class:`.Result` will be considered failed if
     * itself encountered an error while fetching records
     * another result within the same transaction encountered an error while
       fetching records

    .. versionadded: 5.14
    """


# DriverError > ResultError > ResultConsumedError
class ResultConsumedError(ResultError):
    """Raised when trying to access records of a consumed result."""


# DriverError > ResultError > ResultNotSingleError
class ResultNotSingleError(ResultError):
    """Raised when a result should have exactly one record but does not."""


# DriverError > BrokenRecordError
class BrokenRecordError(DriverError):
    """
    Raised when accessing a Record's field that couldn't be decoded.

    This can for instance happen when the server sends a zoned datetime with a
    zone id unknown to the client.
    """


# DriverError > SessionExpired
class SessionExpired(DriverError):
    """
    The session has expired.

    Raised when a session is no longer able to fulfil the purpose described by
    its original parameters.
    """

    def __init__(self, *args: object) -> None:
        super().__init__(*args)
        message = ""
        description = "error: connection exception"
        if len(args) > 0 and isinstance(args[0], str):
            message = args[0]
            description = f"{description}. {message}"
        self._init_gql(
            gql_status="08000",
            message=message,
            description=description,
        )

    def is_retryable(self) -> bool:
        return True


# DriverError > ServiceUnavailable
class ServiceUnavailable(DriverError):
    """
    Raised when no database service is available.

    This may be due to incorrect configuration or could indicate a runtime
    failure of a database service that the driver is unable to route around.
    """

    def __init__(self, *args: object) -> None:
        super().__init__(*args)
        message = ""
        description = "error: connection exception"
        if len(args) > 0 and isinstance(args[0], str):
            message = args[0]
            description = f"{description}. {message}"
        self._init_gql(
            gql_status="08000",
            message=message,
            description=description,
        )

    def is_retryable(self) -> bool:
        return True


# DriverError > ServiceUnavailable > RoutingServiceUnavailable
class RoutingServiceUnavailable(ServiceUnavailable):
    """Raised when no routing service is available."""


# DriverError > ServiceUnavailable > WriteServiceUnavailable
class WriteServiceUnavailable(ServiceUnavailable):
    """Raised when no write service is available."""


# DriverError > ServiceUnavailable > ReadServiceUnavailable
class ReadServiceUnavailable(ServiceUnavailable):
    """Raised when no read service is available."""


# DriverError > ServiceUnavailable > IncompleteCommit
class IncompleteCommit(ServiceUnavailable):
    """
    Raised when the client looses connection while committing a transaction.

    Raised when a disconnection occurs while still waiting for a commit
    response. For non-idempotent write transactions, this leaves the data
    in an unknown state with regard to whether the transaction completed
    successfully or not.
    """

    def __init__(self, *args: object) -> None:
        super().__init__(*args)
        message = ""
        if len(args) > 0 and isinstance(args[0], str):
            message = args[0]
        self._init_gql(
            gql_status="08007",
            message=message,
            description=(
                "error: connection exception - transaction resolution unknown"
            ),
        )

    def is_retryable(self) -> bool:
        return False


# DriverError > ConfigurationError
class ConfigurationError(DriverError):
    """Raised when there is an error concerning a configuration."""


# DriverError > ConfigurationError > AuthConfigurationError
class AuthConfigurationError(ConfigurationError):
    """Raised when there is an error with the authentication configuration."""


# DriverError > ConfigurationError > CertificateConfigurationError
class CertificateConfigurationError(ConfigurationError):
    """Raised when there is an error with the certificate configuration."""


# DriverError > ConfigurationError > UnsupportedServerProduct
class UnsupportedServerProduct(ConfigurationError):
    """
    Raised when an unsupported server product is detected.

    .. versionchanged:: 6.0
        This exception is now a subclass of :class:`ConfigurationError`.
        Before it was a subclass of :class:`Exception`.
    """

    def __init__(self, *args: object) -> None:
        super().__init__(*args)


# DriverError > ConnectionPoolError
class ConnectionPoolError(DriverError):
    """Raised when the connection pool encounters an error."""


# DriverError > ConnectionPoolError > ConnectionAcquisitionTimeoutError
class ConnectionAcquisitionTimeoutError(ConnectionPoolError):
    """
    Raised when no connection became available in time.

    The amount of time is determined by the connection acquisition timeout
    configuration option.
    """



# Copyright (c) "Neo4j"
# Neo4j Sweden AB [https://neo4j.com]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import asyncio
import logging
import socket

from ..._addressing import (
    Address,
    ResolvedAddress,
)
from ...exceptions import ServiceUnavailable
from ..util import AsyncUtil


log = logging.getLogger("neo4j.io")


def _resolved_addresses_from_info(info, host_name):
    resolved = []
    for fam, _, _, _, addr in info:
        if fam == socket.AF_INET6 and addr[3] != 0:
            # skip any IPv6 addresses with a non-zero scope id
            # as these appear to cause problems on some platforms
            continue
        if addr not in resolved:
            resolved.append(addr)
            yield ResolvedAddress(addr, host_name=host_name)


_RETRYABLE_DNS_ERRNOS = {
    socket.EAI_ADDRFAMILY,
    socket.EAI_AGAIN,
    socket.EAI_MEMORY,
    socket.EAI_NODATA,
}


class AsyncNetworkUtil:
    @staticmethod
    async def get_address_info(
        host, port, *, family=0, type=0, proto=0, flags=0
    ):
        loop = asyncio.get_event_loop()
        return await loop.getaddrinfo(
            host, port, family=family, type=type, proto=proto, flags=flags
        )

    @staticmethod
    async def _dns_resolver(address, family=0):
        """
        Apply regular DNS resolution.

        Takes an address object and optional address family for filtering.

        :param address:
        :param family:
        :returns:
        """
        try:
            info = await AsyncNetworkUtil.get_address_info(
                address.host,
                address.port,
                family=family,
                type=socket.SOCK_STREAM,
            )
        except OSError as e:
            if e.errno in _RETRYABLE_DNS_ERRNOS or (
                e.errno == socket.EAI_NONAME
                and (address.host is not None or address.port is not None)
            ):
                raise ServiceUnavailable(
                    f"Failed to DNS resolve address {address}: {e}"
                ) from e
            raise ValueError(
                f"Failed to DNS resolve address {address}: {e}"
            ) from e
        return list(_resolved_addresses_from_info(info, address._host_name))

    @staticmethod
    async def resolve_address(address, family=0, resolver=None):
        """
        Carry out domain name resolution on this Address object.

        If a resolver function is supplied, and is callable, this is
        called first, with this object as its argument. This may yield
        multiple output addresses, which are chained into a subsequent
        regular DNS resolution call. If no resolver function is passed,
        the DNS resolution is carried out on the original Address
        object.

        This function returns a list of resolved Address objects.

        :param address: the Address to resolve
        :param family: optional address family to filter resolved
                       addresses by (e.g. ``socket.AF_INET6``)
        :param resolver: optional customer resolver function to be
                         called before regular DNS resolution
        """
        if isinstance(address, ResolvedAddress):
            yield address
            return

        log.debug("[#0000]  _: <RESOLVE> in: %s", address)
        if resolver:
            addresses_resolved = map(
                Address,
                await AsyncUtil.callback(resolver, address),
            )
            for address_resolved in addresses_resolved:
                log.debug(
                    "[#0000]  _: <RESOLVE> custom resolver out: %s",
                    address_resolved,
                )
                addresses_dns_resolved = await AsyncNetworkUtil._dns_resolver(
                    address_resolved, family=family
                )
                for address_dns_resolved in addresses_dns_resolved:
                    log.debug(
                        "[#0000]  _: <RESOLVE> dns resolver out: %s",
                        address_dns_resolved,
                    )
                    yield address_dns_resolved
        else:
            for address_dns_resolved in await AsyncNetworkUtil._dns_resolver(
                address, family=family
            ):
                log.debug(
                    "[#0000]  _: <RESOLVE> dns resolver out: %s",
                    address_dns_resolved,
                )
                yield address_dns_resolved


class NetworkUtil:
    @staticmethod
    def get_address_info(host, port, *, family=0, type=0, proto=0, flags=0):
        return socket.getaddrinfo(host, port, family, type, proto, flags)

    @staticmethod
    def _dns_resolver(address, family=0):
        """
        Apply regular DNS resolution.

        Takes an address object and optional address family for filtering.

        :param address:
        :param family:
        :returns:
        """
        try:
            info = NetworkUtil.get_address_info(
                address.host,
                address.port,
                family=family,
                type=socket.SOCK_STREAM,
            )
        except OSError as e:
            if e.errno in _RETRYABLE_DNS_ERRNOS or (
                e.errno == socket.EAI_NONAME
                and (address.host is not None or address.port is not None)
            ):
                raise ServiceUnavailable(
                    f"Failed to DNS resolve address {address}: {e}"
                ) from e
            raise ValueError(
                f"Failed to DNS resolve address {address}: {e}"
            ) from e
        return _resolved_addresses_from_info(info, address._host_name)

    @staticmethod
    def resolve_address(address, family=0, resolver=None):
        """
        Carry out domain name resolution on this Address object.

        If a resolver function is supplied, and is callable, this is
        called first, with this object as its argument. This may yield
        multiple output addresses, which are chained into a subsequent
        regular DNS resolution call. If no resolver function is passed,
        the DNS resolution is carried out on the original Address
        object.

        This function returns a list of resolved Address objects.

        :param address: the Address to resolve
        :param family: optional address family to filter resolved
                       addresses by (e.g. ``socket.AF_INET6``)
        :param resolver: optional customer resolver function to be
                         called before regular DNS resolution
        """
        if isinstance(address, ResolvedAddress):
            yield address
            return

        log.debug("[#0000]  _: <RESOLVE> in: %s", address)
        if resolver:
            addresses_resolved = map(Address, resolver(address))
            for address_resolved in addresses_resolved:
                log.debug(
                    "[#0000]  _: <RESOLVE> custom resolver out: %s",
                    address_resolved,
                )
                addresses_dns_resolved = NetworkUtil._dns_resolver(
                    address_resolved, family=family
                )
                for address_dns_resolved in addresses_dns_resolved:
                    log.debug(
                        "[#0000]  _: <RESOLVE> dns resolver out: %s",
                        address_dns_resolved,
                    )
                    yield address_dns_resolved
        else:
            for address_dns_resolved in NetworkUtil._dns_resolver(
                address, family=family
            ):
                log.debug(
                    "[#0000]  _: <RESOLVE> dns resolver out: %s",
                    address_dns_resolved,
                )
                yield address_dns_resolved
